import base64
import csv
import json
import logging
import os
import time
from abc import abstractmethod
from collections.abc import Sequence
from datetime import date, datetime, timedelta
from enum import IntEnum, Enum
from functools import cache
from itertools import islice
from typing import Iterator
from uuid import UUID
from packaging import version

import requests
from lxml import etree
from lxml.etree import QName
from .. import Code

from .. import CFDI, __version__, ResponseError, Signer, Certificate
from . import PAC, Environment, TaxpayerStatus
from ..create.w3.signature import signature_c14n_sha1, _digest, _tobytes
from ..transform import MEXICO_TZ, get_timezone, verify_certificate
from ..utils import iterate

parser = etree.XMLParser(remove_blank_text=True, huge_tree=True)
logger = logging.getLogger(__name__)
current_dir = os.path.dirname(__file__)

LISTADO_COMPLETO_69B_JSON = os.path.join(current_dir, 'Listado_Completo_69-B.json')
REFRESH_TIME = (15 * 86400)  # 15 Days


class EstadoSolicitud(IntEnum):
    Aceptada = 1
    EnProceso = 2
    Terminada = 3
    Error = 4
    Rechazada = 5
    Vencida = 6


class TipoDescargaMasivaTerceros(Enum):
    CFDI = 'CFDI'
    Metadata = 'Metadata'


class TipoDeComprobante(Enum):
    Ingreso = 'I'
    Egreso = 'E'
    Traslado = 'T'
    Nomina = 'N'
    Pago = 'P'


class EstadoComprobante(Enum):
    Cancelado = '0'
    Vigente = '1'


def _service_logger(response_type, response):
    if response["CodEstatus"] == "5000":
        logger.info("%s: %s", response_type, response)
        return
    raise ResponseError(response)


def _certificate_path(certificate_no):
    return f"{certificate_no[0:6]}/{certificate_no[6:12]}/{certificate_no[12:14]}/{certificate_no[14:16]}/{certificate_no[16:18]}/{certificate_no}"


@cache
def _get_listado_69b(refresh_time=REFRESH_TIME):
    try:
        t = os.path.getmtime(LISTADO_COMPLETO_69B_JSON)
    except FileNotFoundError:
        t = -refresh_time

    if time.time() > t + refresh_time:
        r = requests.get(
            url="http://omawww.sat.gob.mx/cifras_sat/Documents/Listado_Completo_69-B.csv",
            headers={
                "User-Agent": __version__.__user_agent__
            }
        )

        if r.status_code == 200:
            lines = str(r.content, 'windows-1250').splitlines(keepends=True)
            cvs_reader = csv.reader(islice(lines, 3, None), delimiter=',', quotechar='"')
            res = {row[1]: row[3] for row in cvs_reader}
            _save_listado_69b(res)
            return res

        logger.error("Unable to get latest Listado 69B, status code: %s", r.status_code)

    if t > 0:
        return _load_listado_69b()
    else:
        raise ResponseError("Unable to load Listado Completo 69B")


def _save_listado_69b(data):
    with open(LISTADO_COMPLETO_69B_JSON, "w", encoding="utf-8") as f:
        json.dump(data, f, ensure_ascii=False, separators=(',', ':'), check_circular=False)


def _load_listado_69b():
    with open(LISTADO_COMPLETO_69B_JSON, "r", encoding="utf-8") as f:
        return json.load(f)


# SAT REQUESTS #
def _set_arguments(element, arguments: dict) -> etree.Element:
    for key, value in arguments.items():
        if value is None:
            continue
        elif isinstance(value, str):
            element.set(key, value)
        elif isinstance(value, UUID):
            element.set(key, str(value))
        elif isinstance(value, datetime):
            element.set(key, value.isoformat(timespec="seconds"))
        elif isinstance(value, date):
            element.set(key, value.isoformat())
        elif isinstance(value, Enum):
            element.set(key, value.value)
        elif isinstance(value, list):
            sub_elem = etree.SubElement(element, etree.QName(element.nsmap[element.prefix], key))
            for (k, v) in value:
                etree.SubElement(sub_elem, etree.QName(element.nsmap[element.prefix], k)).text = v
        else:
            raise ValueError("Invalid Arguments", key, value)


def _get_template(template_name: str) -> etree.Element:
    return etree.parse(
        source=os.path.join(current_dir, "sat_templates", template_name),
        parser=parser
    ).getroot()


class _SATRequest:
    xml_name = None
    soap_url = None
    soap_action = None
    solicitud_xpath = None
    sign_payload = True
    https_verify = True

    def __init__(self, signer: Signer = None, arguments: dict = None):
        self.signer = signer
        self.arguments = arguments or {}

    def get_payload(self):
        template = _get_template(self.xml_name)
        self._prepare_payload(template)
        # return _tobytes(template)
        return etree.tostring(template, encoding="UTF-8")

    @abstractmethod
    def process_response(self, response: etree.Element):
        pass

    def _prepare_payload(self, root: etree.Element):
        sol = root.find(self.solicitud_xpath)
        _set_arguments(sol, self.arguments)
        if self.sign_payload:
            sol.append(
                signature_c14n_sha1(
                    signer=self.signer,
                    element=sol.getparent()
                ).to_xml()
            )


class _CFDIAutenticacion(_SATRequest):
    xml_name = 'autentica.xml'
    soap_url = 'https://cfdidescargamasivasolicitud.clouda.sat.gob.mx/Autenticacion/Autenticacion.svc'
    wsdl_url = 'https://cfdidescargamasivasolicitud.clouda.sat.gob.mx/Autenticacion/Autenticacion.svc?wsdl'
    soap_action = 'http://DescargaMasivaTerceros.gob.mx/IAutenticacion/Autentica'

    DATE_TIME_FORMAT: str = '%Y-%m-%dT%H:%M:%S.%fZ'

    def _prepare_payload(self, root):
        date_created = datetime.utcnow()
        date_expires = date_created + timedelta(seconds=self.arguments.get("seconds", 300))
        security = root.find('{*}Header/{*}Security')

        security.find('{*}Timestamp/{*}Created').text = date_created.strftime(
            self.DATE_TIME_FORMAT)
        security.find('{*}Timestamp/{*}Expires').text = date_expires.strftime(
            self.DATE_TIME_FORMAT)
        security.find('{*}BinarySecurityToken').text = self.signer.certificate_base64()

        security.find('{*}Signature/{*}SignedInfo/{*}Reference/{*}DigestValue').text = _digest(
            security.find('{*}Timestamp')
        )
        security.find('{*}Signature/{*}SignatureValue').text = \
            self.signer.sign_sha1(
                _tobytes(security.find('{*}Signature/{*}SignedInfo'))
            )

    def process_response(self, response):
        authres = response.find('{*}Body/{*}AutenticaResponse/{*}AutenticaResult')
        timestamps = response.find('{*}Header/{*}Security/{*}Timestamp')

        def todatetime(name):
            return datetime.fromisoformat(timestamps.find(name).text[:-1])

        return {
            "Created": todatetime('{*}Created'),
            "Expires": todatetime('{*}Expires'),
            "AutenticaResult": authres.text
        }


class _CFDISolicitaDescarga(_SATRequest):
    xml_name = 'solicita.xml'
    soap_url = 'https://cfdidescargamasivasolicitud.clouda.sat.gob.mx/SolicitaDescargaService.svc'
    soap_action = 'http://DescargaMasivaTerceros.sat.gob.mx/ISolicitaDescargaService/SolicitaDescarga'
    solicitud_xpath = '{*}Body/{*}SolicitaDescarga/{*}solicitud'

    def process_response(self, response: etree.Element):
        res = response.find('{*}Body/{*}SolicitaDescargaResponse/{*}SolicitaDescargaResult')
        return {
            **res.attrib
        }


class _CFDIVerificaSolicitudDescarga(_SATRequest):
    xml_name = 'verifica.xml'
    soap_url = 'https://cfdidescargamasivasolicitud.clouda.sat.gob.mx/VerificaSolicitudDescargaService.svc'
    soap_action = 'http://DescargaMasivaTerceros.sat.gob.mx/IVerificaSolicitudDescargaService/VerificaSolicitudDescarga'
    solicitud_xpath = '{*}Body/{*}VerificaSolicitudDescarga/{*}solicitud'

    def process_response(self, response: etree.Element):
        res = response.find('{*}Body/{*}VerificaSolicitudDescargaResponse/{*}VerificaSolicitudDescargaResult')

        def descarga_result(node):
            result = dict()
            result['IdsPaquetes'] = [n.text for n in node.iterfind('{http://DescargaMasivaTerceros.sat.gob.mx}IdsPaquetes')]
            at = node.attrib.get('CodEstatus')
            if at is not None:
                result['CodEstatus'] = at
            estado_solicitud = int(node.attrib['EstadoSolicitud'])
            result['EstadoSolicitud'] = Code(estado_solicitud, EstadoSolicitud(estado_solicitud).name)
            at = node.attrib.get('CodigoEstadoSolicitud')
            if at is not None:
                result['CodigoEstadoSolicitud'] = at
            result['NumeroCFDIs'] = int(node.attrib['NumeroCFDIs'])
            at = node.attrib.get('Mensaje')
            if at is not None:
                result['Mensaje'] = at
            return result

        return descarga_result(res)


class _CFDIDescargaMasiva(_SATRequest):
    xml_name = 'descarga.xml'
    soap_url = 'https://cfdidescargamasiva.clouda.sat.gob.mx/DescargaMasivaService.svc'
    soap_action = 'http://DescargaMasivaTerceros.sat.gob.mx/IDescargaMasivaTercerosService/Descargar'
    solicitud_xpath = '{*}Body/{*}PeticionDescargaMasivaTercerosEntrada/{*}peticionDescarga'

    def process_response(self, response: etree.Element):
        paquete = response.find('{*}Body/{*}RespuestaDescargaMasivaTercerosSalida/{*}Paquete')
        header = response.find('{*}Header/{*}respuesta')

        def respuesta(node):
            result = dict()
            at = node.attrib.get('CodEstatus')
            if at is not None:
                result['CodEstatus'] = at
            at = node.attrib.get('Mensaje')
            if at is not None:
                result['Mensaje'] = at
            return result

        return respuesta(header), paquete.text


class _RetenAutenticacion(_CFDIAutenticacion):
    soap_url = 'https://retendescargamasivasolicitud.clouda.sat.gob.mx/Autenticacion/Autenticacion.svc'


class _RetenSolicitaDescarga(_CFDISolicitaDescarga):
    soap_url = 'https://retendescargamasivasolicitud.clouda.sat.gob.mx/SolicitaDescargaService.svc'


class _RetenVerificaSolicitudDescarga(_CFDIVerificaSolicitudDescarga):
    soap_url = 'https://retendescargamasivasolicitud.clouda.sat.gob.mx/VerificaSolicitudDescargaService.svc'


class _RetenDescargaMasiva(_CFDIDescargaMasiva):
    soap_url = 'https://retendescargamasiva.clouda.sat.gob.mx/DescargaMasivaService.svc'


class SAT(PAC):
    RFC = "SAT970701NN3"

    def __init__(self, signer: Signer = None, environment=Environment.PRODUCTION):
        self.signer = signer
        super().__init__(environment)
        self.token_comprobante = None
        self.token_retencion = None
        self.wait_time = 60

    def _get_headers(self, soap_action, needs_token_fn) -> dict:
        headers = {
            'Content-type': 'text/xml;charset="utf-8"',
            'Accept': 'text/xml',
            'Cache-Control': 'no-cache',
            "User-Agent": __version__.__user_agent__,
            'SOAPAction': soap_action
        }
        if needs_token_fn:
            headers['Authorization'] = f'WRAP access_token="{needs_token_fn()}"'
        return headers

    def _request(self, soap_url, data, soap_action, needs_token_fn, verify=True):
        response = requests.post(
            url=soap_url,
            data=data,
            headers=self._get_headers(soap_action, needs_token_fn=needs_token_fn),
            verify=verify
        )

        if not response.ok:
            raise ResponseError(response)

        return etree.fromstring(
            response.content,
            parser=parser
        )

    def _get_token_comprobante(self):
        if self.token_comprobante is None or self.token_comprobante["Expires"] <= datetime.utcnow() + timedelta(seconds=30):
            self.token_comprobante = self._autentica_comprobante()
        return self.token_comprobante["AutenticaResult"]

    def _get_token_retencion(self):
        if self.token_retencion is None or self.token_retencion["Expires"] <= datetime.utcnow() + timedelta(seconds=30):
            self.token_retencion = self._autentica_retencion()
        return self.token_retencion["AutenticaResult"]

    def _autentica_comprobante(self):
        return self._execute_req(
            _CFDIAutenticacion(signer=self.signer),
            needs_token_fn=None
        )

    def _autentica_retencion(self):
        return self._execute_req(
            _RetenAutenticacion(signer=self.signer),
            needs_token_fn=None
        )

    def _execute_req(self, req: _SATRequest, needs_token_fn):
        payload = req.get_payload()

        xml = self._request(
            soap_url=req.soap_url,
            data=payload,
            soap_action=req.soap_action,
            verify=req.https_verify,
            needs_token_fn=needs_token_fn
        )

        return req.process_response(xml)

    def status(self, cfdi: CFDI) -> dict:
        rfc_emisor = cfdi["Emisor"]["Rfc"]
        rfc_receptor = cfdi["Receptor"]["Rfc"]
        total = cfdi["Total"]
        uuid = cfdi["Complemento"]["TimbreFiscalDigital"]["UUID"]

        template = f'<Envelope xmlns="http://schemas.xmlsoap.org/soap/envelope/" xmlns:tem="http://tempuri.org/"><Body><tem:Consulta><tem:expresionImpresa>' \
                   f'<![CDATA[?re={rfc_emisor}&rr={rfc_receptor}&tt={total}&id={uuid}]]>' \
                   f'</tem:expresionImpresa></tem:Consulta></Body></Envelope>'

        match self.environment:
            case Environment.PRODUCTION:
                host = "https://consultaqr.facturaelectronica.sat.gob.mx"
                verify = True
            case Environment.TEST:
                host = "https://pruebacfdiconsultaqr.cloudapp.net"
                verify = False
            case _:
                raise NotImplementedError("Environment not Supported")

        res = self._request(
            soap_url=f'{host}/ConsultaCFDIService.svc',
            data=template,
            soap_action='http://tempuri.org/IConsultaCFDIService/Consulta',
            verify=verify,
            needs_token_fn=None
        )

        res = res.find('{*}Body/{*}ConsultaResponse/{*}ConsultaResult')
        return {
            QName(i.tag).localname: i.text
            for i in res
        }

    def validate(self, cfdi: CFDI):
        """
                verify if the CFDI is valid based on its signatures and certificates
                this method might need improvements
                :return:
                True is all certificate and signatures verifications passed
                """
        match cfdi.tag:
            case '{http://www.sat.gob.mx/cfd/3}Comprobante' | '{http://www.sat.gob.mx/cfd/4}Comprobante':
                return self._validate(
                    cfdi,
                    certificate=cfdi["Certificado"],
                    num_certificate=cfdi['NoCertificado'],
                    rfc=cfdi["Emisor"]["Rfc"],
                    seal=cfdi["Sello"],
                    date=cfdi['Fecha'],
                    postal_code=cfdi['LugarExpedicion'],
                    stamp_version="1.0" if version.parse(cfdi["Version"]) <= version.parse("3.2") else "1.1"
                )

            case '{http://www.sat.gob.mx/esquemas/retencionpago/1}Retenciones':
                return self._validate(
                    cfdi,
                    certificate=cfdi["Cert"],
                    num_certificate=cfdi['NumCert'],
                    rfc=cfdi["Emisor"]["RFCEmisor"],
                    seal=cfdi["Sello"],
                    date=cfdi['FechaExp'],
                    postal_code=None,
                    stamp_version="1.0"
                )

            case '{http://www.sat.gob.mx/esquemas/retencionpago/2}Retenciones':
                return self._validate(
                    cfdi,
                    certificate=cfdi["Certificado"],
                    num_certificate=cfdi['NoCertificado'],
                    rfc=cfdi["Emisor"]["RfcE"],
                    seal=cfdi["Sello"],
                    date=cfdi['FechaExp'],
                    postal_code=cfdi['LugarExpRetenc'],
                    stamp_version="1.1"
                )

    def _validate(self, cfdi, certificate, num_certificate, rfc, seal, date: datetime, postal_code, stamp_version):
        if not date.tzinfo:
            # debe corresponder con la hora local donde se expide el comprobante
            date = date.replace(tzinfo=get_timezone(postal_code))

        if stamp_version == "1.0":
            verify_fn = Certificate.verify_sha1
        else:
            verify_fn = Certificate.verify_sha256

        cert = Certificate.load_certificate(
            base64.b64decode(certificate)
        )

        if not verify_certificate(cert, at=date):
            return False

        if not cert.certificate_number == num_certificate:
            return False

        if not cert.rfc == rfc:
            return False

        if not verify_fn(
                self=cert,
                data=cfdi.cadena_original().encode(),
                signature=base64.b64decode(seal)
        ):
            return False

        # Timbre
        timbre = cfdi["Complemento"]["TimbreFiscalDigital"]
        # debe corresponder con la hora de la Zona Centro del Sistema de Horario en México.
        stamp_date = timbre["FechaTimbrado"].replace(tzinfo=MEXICO_TZ)

        if not timbre["Version"] == stamp_version:
            return False

        if not timbre["SelloCFD"] == seal:
            return False

        cert_sat = self.recover_certificate(
            no_certificado=timbre["NoCertificadoSAT"]
        )

        if rfc_pac := timbre.get("RfcProvCertif"):
            if not rfc_pac == cert_sat.rfc_pac:
                return False

        if not verify_certificate(cert, at=stamp_date):
            return False

        if not verify_fn(
                self=cert_sat,
                data=timbre.cadena_original().encode(),
                signature=base64.b64decode(timbre["SelloSAT"])
        ):
            return False

        # El rango de la fecha de generación no debe de ser mayor a 72 horas para la emisión del timbre.
        if not date <= stamp_date + timedelta(hours=72):
            return False

        return True

    def list_69b(self, rfc: str) -> TaxpayerStatus | None:
        listado = _get_listado_69b()
        r = listado.get(rfc)
        if r:
            return TaxpayerStatus(r)
        return None

    def recover_comprobante_request(
            self,
            fecha_inicial: date = None,
            fecha_final: date = None,
            rfc_receptor: str | Sequence[str] = None,
            rfc_emisor: str = None,
            tipo_solicitud: TipoDescargaMasivaTerceros | str = TipoDescargaMasivaTerceros.CFDI,
            tipo_comprobante: TipoDeComprobante | str = None,
            estado_comprobante: EstadoComprobante | str = None,
            rfc_a_cuenta_terceros: str = None,
            complemento: str = None,
            uuid: str | UUID = None) -> dict:
        """
        Esta operación permite solicitar la descarga de CFDIs o Metadata y como
        resultado devuelve un id de solicitud o estatus de la petición realizada.

        :param fecha_inicial: Solo se buscarán CFDI, cuya fecha de emisión sea igual o mayor a la fecha inicial indicada en este parámetro.
            Parámetro obligatorio. Este parámetro no debe declararse en caso de realizar una consulta por el folio fiscal (UUID).
        :param fecha_final: Solo se buscarán CFDI, cuya fecha de emisión sea igual o menor a la fecha final indicada en este parámetro.
            Parámetro obligatorio. Este parámetro no debe declararse en caso de realizar una consulta por el folio fiscal (UUID).
        :param rfc_receptor: Contiene el/los RFCs receptores de los cuales se quiere consultar los CFDIs
            Importante: El campo RfcReceptor, únicamente permite la captura de 5 registros como máximo
        :param rfc_emisor: Contiene el RFC del emisor del cual se quiere consultar los CFDI.
            Parámetro obligatorio. Este parámetro no debe declararse en caso de realizar una consulta por el folio fiscal (UUID).
        :param tipo_solicitud: Define el tipo de descarga
        :param tipo_comprobante: Define el tipo de comprobante
        :param estado_comprobante: Define el estado del comprobante
        :param rfc_a_cuenta_terceros: Contiene el RFC del a cuenta a tercero del cual se quiere consultar los CFDIs
        :param complemento: Define el complemento de CFDI a descargar
        :param uuid: Folio Fiscal
        :return: respuesta de solicitud de descarga
        """
        arguments = {
            'FechaFinal': fecha_final,
            'FechaInicial': fecha_inicial,
            'RfcEmisor': rfc_emisor,
            'RfcReceptores': [('RfcReceptor', r) for r in iterate(rfc_receptor)],
            'RfcSolicitante': self.signer.rfc,
            'TipoSolicitud': tipo_solicitud,
            'TipoComprobante': tipo_comprobante,
            'EstadoComprobante': estado_comprobante,
            'RfcACuentaTerceros': rfc_a_cuenta_terceros,
            'Complemento': complemento,
            'UUID': uuid,
        }

        return self._execute_req(
            _CFDISolicitaDescarga(
                signer=self.signer,
                arguments=arguments
            ),
            needs_token_fn=self._get_token_comprobante
        )

    def recover_comprobante_status(self, id_solicitud: str) -> dict:
        return self._execute_req(
            _CFDIVerificaSolicitudDescarga(
                signer=self.signer,
                arguments={
                    'RfcSolicitante': self.signer.rfc,
                    'IdSolicitud': id_solicitud,
                }
            ),
            needs_token_fn=self._get_token_comprobante
        )

    def recover_comprobante_download(self, id_paquete: str) -> (dict, str):
        return self._execute_req(
            _CFDIDescargaMasiva(
                signer=self.signer,
                arguments={
                    'RfcSolicitante': self.signer.rfc,
                    'IdPaquete': id_paquete,
                }
            ),
            needs_token_fn=self._get_token_comprobante
        )

    def recover_comprobante_iwait(
            self,
            fecha_inicial: date = None,
            fecha_final: date = None,
            rfc_receptor: str | Sequence[str] = None,
            rfc_emisor: str = None,
            tipo_solicitud: TipoDescargaMasivaTerceros | str = TipoDescargaMasivaTerceros.CFDI,
            tipo_comprobante: TipoDeComprobante | str = None,
            estado_comprobante: EstadoComprobante | str = None,
            rfc_a_cuenta_terceros: str = None,
            complemento: str = None,
            uuid: str | UUID = None,
            id_solicitud: str | UUID = None
    ) -> Iterator[tuple[str, bytes]]:
        """
        Itera sobre los paquetes obtenidos

        :param fecha_inicial: Solo se buscarán CFDI, cuya fecha de emisión sea igual o mayor a la fecha inicial indicada en este parámetro.
            Parámetro obligatorio. Este parámetro no debe declararse en caso de realizar una consulta por el folio fiscal (UUID).
        :param fecha_final: Solo se buscarán CFDI, cuya fecha de emisión sea igual o menor a la fecha final indicada en este parámetro.
            Parámetro obligatorio. Este parámetro no debe declararse en caso de realizar una consulta por el folio fiscal (UUID).
        :param rfc_receptor: Contiene el/los RFCs receptores de los cuales se quiere consultar los CFDIs
            Importante: El campo RfcReceptor, únicamente permite la captura de 5 registros como máximo
        :param rfc_emisor: Contiene el RFC del emisor del cual se quiere consultar los CFDI.
            Parámetro obligatorio. Este parámetro no debe declararse en caso de realizar una consulta por el folio fiscal (UUID).
        :param tipo_solicitud: Define el tipo de descarga
        :param tipo_comprobante: Define el tipo de comprobante
        :param estado_comprobante: Define el estado del comprobante
        :param rfc_a_cuenta_terceros: Contiene el RFC del a cuenta a tercero del cual se quiere consultar los CFDIs
        :param complemento: Define el complemento de CFDI a descargar
        :param uuid: Folio Fiscal
        :param id_solicitud: Si ya se cuenta con el id de una solicitud anterior, solo pasar este parametro
        :return: Iterador sobre packetes y su contenido en bytes
        """
        if not id_solicitud:
            response = self.recover_comprobante_request(
                fecha_inicial=fecha_inicial,
                fecha_final=fecha_final,
                rfc_receptor=rfc_receptor,
                rfc_emisor=rfc_emisor,
                tipo_solicitud=tipo_solicitud,
                tipo_comprobante=tipo_comprobante,
                estado_comprobante=estado_comprobante,
                rfc_a_cuenta_terceros=rfc_a_cuenta_terceros,
                complemento=complemento,
                uuid=uuid
            )
            _service_logger("SolicitaDescargaResult", response)
            id_solicitud = response['IdSolicitud']
            time.sleep(self.wait_time)

        while True:
            response = self.recover_comprobante_status(
                id_solicitud=id_solicitud
            )
            _service_logger("VerificaSolicitudDescargaResult", response)
            est = response["EstadoSolicitud"]
            if est == EstadoSolicitud.Terminada:
                for id_paquete in response['IdsPaquetes']:
                    response, paquete = self.recover_comprobante_download(
                        id_paquete=id_paquete
                    )
                    _service_logger("RespuestaDescargaMasivaTercerosSalida", response)
                    yield id_paquete, base64.b64decode(paquete)
                break
            if est in [EstadoSolicitud.Aceptada, EstadoSolicitud.EnProceso]:
                time.sleep(self.wait_time)
                continue
            break

    def recover_retencion_request(
            self,
            fecha_inicial: date = None,
            fecha_final: date = None,
            rfc_receptor: str | Sequence[str] = None,
            rfc_emisor: str = None,
            tipo_solicitud: TipoDescargaMasivaTerceros | str = TipoDescargaMasivaTerceros.CFDI,
            tipo_comprobante: TipoDeComprobante | str = None,
            estado_comprobante: EstadoComprobante | str = None,
            rfc_a_cuenta_terceros: str = None,
            complemento: str = None,
            uuid: str | UUID = None) -> dict:
        """
        Esta operación permite solicitar la descarga de CFDIs o Metadata y como
        resultado devuelve un id de solicitud o estatus de la petición realizada.

        :param fecha_inicial: Solo se buscarán CFDI, cuya fecha de emisión sea igual o mayor a la fecha inicial indicada en este parámetro.
            Parámetro obligatorio. Este parámetro no debe declararse en caso de realizar una consulta por el folio fiscal (UUID).
        :param fecha_final: Solo se buscarán CFDI, cuya fecha de emisión sea igual o menor a la fecha final indicada en este parámetro.
            Parámetro obligatorio. Este parámetro no debe declararse en caso de realizar una consulta por el folio fiscal (UUID).
        :param rfc_receptor: Contiene el/los RFCs receptores de los cuales se quiere consultar los CFDIs
            Importante: El campo RfcReceptor, únicamente permite la captura de 5 registros como máximo
        :param rfc_emisor: Contiene el RFC del emisor del cual se quiere consultar los CFDI.
            Parámetro obligatorio. Este parámetro no debe declararse en caso de realizar una consulta por el folio fiscal (UUID).
        :param tipo_solicitud: Define el tipo de descarga
        :param tipo_comprobante: Define el tipo de comprobante
        :param estado_comprobante: Define el estado del comprobante
        :param rfc_a_cuenta_terceros: Contiene el RFC del a cuenta a tercero del cual se quiere consultar los CFDIs
        :param complemento: Define el complemento de CFDI a descargar
        :param uuid: Folio Fiscal
        :return: respuesta de solicitud de descarga
        """
        arguments = {
            'FechaFinal': fecha_final,
            'FechaInicial': fecha_inicial,
            'RfcEmisor': rfc_emisor,
            'RfcReceptores': [('RfcReceptor', r) for r in iterate(rfc_receptor)],
            'RfcSolicitante': self.signer.rfc,
            'TipoSolicitud': tipo_solicitud,
            'TipoComprobante': tipo_comprobante,
            'EstadoComprobante': estado_comprobante,
            'RfcACuentaTerceros': rfc_a_cuenta_terceros,
            'Complemento': complemento,
            'UUID': uuid,
        }

        return self._execute_req(
            _RetenSolicitaDescarga(
                signer=self.signer,
                arguments=arguments
            ),
            needs_token_fn=self._get_token_retencion
        )

    def recover_retencion_status(self, id_solicitud: str) -> dict:
        return self._execute_req(
            _RetenVerificaSolicitudDescarga(
                signer=self.signer,
                arguments={
                    'RfcSolicitante': self.signer.rfc,
                    'IdSolicitud': id_solicitud,
                }
            ),
            needs_token_fn=self._get_token_retencion
        )

    def recover_retencion_download(self, id_paquete: str) -> (dict, str):
        return self._execute_req(
            _RetenDescargaMasiva(
                signer=self.signer,
                arguments={
                    'RfcSolicitante': self.signer.rfc,
                    'IdPaquete': id_paquete,
                }
            ),
            needs_token_fn=self._get_token_retencion
        )

    def recover_retencion_iwait(
            self,
            fecha_inicial: date = None,
            fecha_final: date = None,
            rfc_receptor: str | Sequence[str] = None,
            rfc_emisor: str = None,
            tipo_solicitud: TipoDescargaMasivaTerceros | str = TipoDescargaMasivaTerceros.CFDI,
            tipo_comprobante: TipoDeComprobante | str = None,
            estado_comprobante: EstadoComprobante | str = None,
            rfc_a_cuenta_terceros: str = None,
            complemento: str = None,
            uuid: str | UUID = None,
            id_solicitud: str | UUID = None
    ) -> Iterator[tuple[str, bytes]]:
        """
        Itera sobre los paquetes obtenidos

        :param fecha_inicial: Solo se buscarán CFDI, cuya fecha de emisión sea igual o mayor a la fecha inicial indicada en este parámetro.
            Parámetro obligatorio. Este parámetro no debe declararse en caso de realizar una consulta por el folio fiscal (UUID).
        :param fecha_final: Solo se buscarán CFDI, cuya fecha de emisión sea igual o menor a la fecha final indicada en este parámetro.
            Parámetro obligatorio. Este parámetro no debe declararse en caso de realizar una consulta por el folio fiscal (UUID).
        :param rfc_receptor: Contiene el/los RFCs receptores de los cuales se quiere consultar los CFDIs
            Importante: El campo RfcReceptor, únicamente permite la captura de 5 registros como máximo
        :param rfc_emisor: Contiene el RFC del emisor del cual se quiere consultar los CFDI.
            Parámetro obligatorio. Este parámetro no debe declararse en caso de realizar una consulta por el folio fiscal (UUID).
        :param tipo_solicitud: Define el tipo de descarga
        :param tipo_comprobante: Define el tipo de comprobante
        :param estado_comprobante: Define el estado del comprobante
        :param rfc_a_cuenta_terceros: Contiene el RFC del a cuenta a tercero del cual se quiere consultar los CFDIs
        :param complemento: Define el complemento de CFDI a descargar
        :param uuid: Folio Fiscal
        :param id_solicitud: Si ya se cuenta con el id de una solicitud anterior, solo pasar este parametro
        :return: Iterador sobre packetes y su contenido en bytes
        """
        if not id_solicitud:
            response = self.recover_retencion_request(
                fecha_inicial=fecha_inicial,
                fecha_final=fecha_final,
                rfc_receptor=rfc_receptor,
                rfc_emisor=rfc_emisor,
                tipo_solicitud=tipo_solicitud,
                tipo_comprobante=tipo_comprobante,
                estado_comprobante=estado_comprobante,
                rfc_a_cuenta_terceros=rfc_a_cuenta_terceros,
                complemento=complemento,
                uuid=uuid
            )
            _service_logger("SolicitaDescargaResult", response)
            id_solicitud = response['IdSolicitud']
            time.sleep(self.wait_time)

        while True:
            response = self.recover_retencion_status(
                id_solicitud=id_solicitud
            )
            _service_logger("VerificaSolicitudDescargaResult", response)
            est = response["EstadoSolicitud"]
            if est == EstadoSolicitud.Terminada:
                for id_paquete in response['IdsPaquetes']:
                    response, paquete = self.recover_retencion_download(
                        id_paquete=id_paquete
                    )
                    _service_logger("RespuestaDescargaMasivaTercerosSalida", response)
                    yield id_paquete, base64.b64decode(paquete)
                break
            if est in [EstadoSolicitud.Aceptada, EstadoSolicitud.EnProceso]:
                time.sleep(self.wait_time)
                continue
            break

    def recover_certificate(self, no_certificado: str | int) -> Certificate:
        path = _certificate_path(no_certificado)
        file_name = os.path.join(current_dir, f"rdc.sat.gob.mx/rccf/{no_certificado}.cer")

        if os.path.exists(file_name):
            with open(file_name, "rb") as f:
                cert_data = f.read()
        else:
            r = requests.get(
                url=f"https://rdc.sat.gob.mx/rccf/{path}.cer",
                headers={
                    "User-Agent": __version__.__user_agent__
                }
            )
            if r.status_code == 200:
                cert_data = r.content

                os.makedirs(os.path.dirname(file_name), exist_ok=True)
                with open(file_name, 'wb') as f:
                    f.write(cert_data)
            else:
                raise ResponseError(r)

        return Certificate.load_certificate(cert_data)



from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ... import CFDI, XElement, ScalarMap


class SelloDigitalContElec:
    """
    Documento requerido para el Sello Digital del SAT que da constancia de recibo del archivo de la contabilidad electrónica.
    """
    tag = '{www.sat.gob.mx/esquemas/ContabilidadE/1_3/SelloDigitalContElec}SelloDigitalContElec'
    version = '1.3'
    
    def __new__(
            cls,
            folio: str,
            rfc: str,
            fechade_sello: datetime,
            no_certificado_sat: str,
            sello_sat: str,
            sello: str = None,
    ) -> CFDI: 
        """
        Documento requerido para el Sello Digital del SAT que da constancia de recibo del archivo de la contabilidad electrónica.
        
        :param folio: Atributo requerido para expresar los 22 caracteres del folio asignado por el SAT en la recepción de los archivos.
        :param rfc: Atributo requerido para expresar el RFC del contribuyente que envía los datos
        :param fechade_sello: Atributo requerido para expresar la fecha y hora de la generación del Sello digital del SAT. Se expresa en la forma aaaa-mm-ddThh:mm:ss, de acuerdo con la especificación ISO 8601
        :param no_certificado_sat: Atributo requerido para expresar el número de serie del certificado del SAT usado para generar el sello digital.
        :param sello_sat: Atributo requerido para contener el sello digital del SAT. El sello deberá ser expresado cómo una cadena de texto en formato Base 64.
        :param sello: Atributo opcional para contener el sello digital del archivo de contabilidad electrónica, que corresponda. El sello deberá ser expresado cómo una cadena de texto en formato Base 64.
        :return: objeto CFDI
        """
        
        cfdi = CFDI({
            'Version': cls.version,
            'Folio': folio,
            'RFC': rfc,
            'FechadeSello': fechade_sello,
            'NoCertificadoSAT': no_certificado_sat,
            'SelloSAT': sello_sat,
            'Sello': sello,
        })
        cfdi.tag = cls.tag
        return cfdi

        




from collections.abc import *
from datetime import datetime
from decimal import Decimal

from .. import Signer
from ... import CFDI, XElement, ScalarMap
from ...transform import get_timezone


class CfdiRetenRelacionados(ScalarMap):
    """
    http://www.sat.gob.mx/esquemas/retencionpago/2
    Nodo opcional para precisar la información de los comprobantes relacionados.
    """

    def __init__(
            self,
            tipo_relacion: str,
            uuid: str,
    ):
        """
        Nodo opcional para precisar la información de los comprobantes relacionados.
        
        :param tipo_relacion: Atributo requerido para indicar la clave de la relación que existe entre éste que se está generando y el comprobante que ampara retenciones e información de pagos previos.
        :param uuid: Atributo requerido para registrar el folio fiscal (UUID) de un comprobante que ampara retención e información de pagos, relacionado con el presente comprobante, ejemplo: Si éste sustituye a un comprobante cancelado.
        """

        super().__init__({
            'TipoRelacion': tipo_relacion,
            'UUID': uuid,
        })


class ImpRetenidos(ScalarMap):
    """
    http://www.sat.gob.mx/esquemas/retencionpago/2
    Nodo opcional para expresar el total de los impuestos retenidos que se desprenden de los conceptos expresados en el comprobante que ampara retenciones e información de pagos.
    """

    def __init__(
            self,
            monto_ret: Decimal | int,
            tipo_pago_ret: str,
            base_ret: Decimal | int = None,
            impuesto_ret: str = None,
    ):
        """
        Nodo opcional para expresar el total de los impuestos retenidos que se desprenden de los conceptos expresados en el comprobante que ampara retenciones e información de pagos.
        
        :param monto_ret: Atributo requerido para expresar el importe del impuesto retenido en el período o ejercicio que se relaciona en el comprobante que ampara retenciones e información de pagos. No se permiten valores negativos.
        :param tipo_pago_ret: Atributo requerido para precisar la clave del tipo del efecto que se le da al monto de la retención.
        :param base_ret: Atributo opcional para expresar la base del impuesto, que puede ser la diferencia entre los ingresos percibidos y las deducciones autorizadas. No se permiten valores negativos.
        :param impuesto_ret: Atributo opcional para señalar el tipo de impuesto retenido del período o ejercicio conforme al catálogo.
        """

        super().__init__({
            'MontoRet': monto_ret,
            'TipoPagoRet': tipo_pago_ret,
            'BaseRet': base_ret,
            'ImpuestoRet': impuesto_ret,
        })


class Totales(ScalarMap):
    """
    http://www.sat.gob.mx/esquemas/retencionpago/2
    Nodo requerido para expresar el total de las retenciones e información de los pagos efectuados en el período que ampara el comprobante.
    """

    def __init__(
            self,
            monto_tot_operacion: Decimal | int,
            monto_tot_grav: Decimal | int,
            monto_tot_exent: Decimal | int,
            monto_tot_ret: Decimal | int,
            utilidad_bimestral: Decimal | int = None,
            isr_correspondiente: Decimal | int = None,
            imp_retenidos: Sequence[ImpRetenidos | dict] = None,
    ):
        """
        Nodo requerido para expresar el total de las retenciones e información de los pagos efectuados en el período que ampara el comprobante.
        
        :param monto_tot_operacion: Atributo requerido para expresar el total del monto de la operación que se relaciona en el comprobante que ampara retenciones e información de pagos. No se permiten valores negativos.
        :param monto_tot_grav: Atributo requerido para expresar el total del monto gravado de la operación que se relaciona en el comprobante que ampara retenciones e información de pagos. No se permiten valores negativos.
        :param monto_tot_exent: Atributo requerido para expresar el total del monto exento de la operación que se relaciona en el comprobante que ampara retenciones e información de pagos. No se permiten valores negativos.
        :param monto_tot_ret: Atributo requerido para expresar el monto total de las retenciones. Sumatoria de los montos de retención del nodo ImpRetenidos. No se permiten valores negativos.
        :param utilidad_bimestral: Atributo condicional para expresar el monto de la utilidad bimestral.
        :param isr_correspondiente: Atributo condicional para expresar el monto del ISR correspondiente al bimestre.
        :param imp_retenidos: Nodo opcional para expresar el total de los impuestos retenidos que se desprenden de los conceptos expresados en el comprobante que ampara retenciones e información de pagos.
        """

        super().__init__({
            'MontoTotOperacion': monto_tot_operacion,
            'MontoTotGrav': monto_tot_grav,
            'MontoTotExent': monto_tot_exent,
            'MontoTotRet': monto_tot_ret,
            'UtilidadBimestral': utilidad_bimestral,
            'ISRCorrespondiente': isr_correspondiente,
            'ImpRetenidos': imp_retenidos,
        })


class Periodo(ScalarMap):
    """
    http://www.sat.gob.mx/esquemas/retencionpago/2
    Nodo requerido para expresar el período que corresponde al comprobante que ampara retenciones e información de pagos.
    """

    def __init__(
            self,
            mes_ini: str,
            mes_fin: str,
            ejercicio: str,
    ):
        """
        Nodo requerido para expresar el período que corresponde al comprobante que ampara retenciones e información de pagos.
        
        :param mes_ini: Atributo requerido para expresar la clave del mes inicial del período de la retención e información de pagos.
        :param mes_fin: Atributo requerido para expresar la clave del mes final del período de la retención e información de pagos.
        :param ejercicio: Atributo requerido para la expresión del ejercicio fiscal (año) de la retención e información de pagos.
        """

        super().__init__({
            'MesIni': mes_ini,
            'MesFin': mes_fin,
            'Ejercicio': ejercicio,
        })


class Extranjero(ScalarMap):
    """
    http://www.sat.gob.mx/esquemas/retencionpago/2
    Nodo requerido para expresar la información del contribuyente receptor del comprobante que ampara retenciones e información de pagos, cuando sea residente en el extranjero.
    """

    def __init__(
            self,
            nom_den_raz_soc_r: str,
            num_reg_id_trib_r: str = None,
    ):
        """
        Nodo requerido para expresar la información del contribuyente receptor del comprobante que ampara retenciones e información de pagos, cuando sea residente en el extranjero.
        
        :param nom_den_raz_soc_r: Atributo requerido para expresar el nombre, denominación o razón social del receptor del comprobante que ampara retenciones e información de pagos, cuando sea residente en el extranjero.
        :param num_reg_id_trib_r: Atributo opcional para expresar el número de registro de identificación fiscal del receptor del comprobante que ampara retenciones e información de pagos, cuando sea residente en el extranjero.
        """

        super().__init__({
            'NomDenRazSocR': nom_den_raz_soc_r,
            'NumRegIdTribR': num_reg_id_trib_r,
        })


class Nacional(ScalarMap):
    """
    http://www.sat.gob.mx/esquemas/retencionpago/2
    Nodo requerido para expresar la información del contribuyente receptor en caso de que sea de nacionalidad mexicana.
    """

    def __init__(
            self,
            rfc_r: str,
            nom_den_raz_soc_r: str,
            domicilio_fiscal_r: str,
            curp_r: str = None,
    ):
        """
        Nodo requerido para expresar la información del contribuyente receptor en caso de que sea de nacionalidad mexicana.
        
        :param rfc_r: Atributo requerido para registrar la clave del Registro Federal de Contribuyentes correspondiente al contribuyente receptor del comprobante que ampara retenciones e información de pagos.
        :param nom_den_raz_soc_r: Atributo requerido para registrar el(los) nombre(s), primer apellido, segundo apellido, según corresponda, denominación o razón social del contribuyente, inscrito en el RFC, del receptor del comprobante que ampara retenciones e información de pagos.
        :param domicilio_fiscal_r: Atributo requerido para registrar el código postal del domicilio fiscal del receptor del comprobante que ampara retenciones e información de pagos.
        :param curp_r: Atributo opcional para la Clave Única del Registro Poblacional del contribuyente receptor del comprobante que ampara retenciones e información de pagos.
        """

        super().__init__({
            'RfcR': rfc_r,
            'NomDenRazSocR': nom_den_raz_soc_r,
            'DomicilioFiscalR': domicilio_fiscal_r,
            'CurpR': curp_r,
        })


class Receptor(ScalarMap):
    """
    http://www.sat.gob.mx/esquemas/retencionpago/2
    Nodo requerido para expresar la información del contribuyente receptor del comprobante que ampara retenciones e información de pagos.
    """

    def __init__(
            self,
            nacionalidad_r: str,
            nacional: Nacional | dict = None,
            extranjero: Extranjero | dict = None,
    ):
        """
        Nodo requerido para expresar la información del contribuyente receptor del comprobante que ampara retenciones e información de pagos.
        
        :param nacionalidad_r: Atributo requerido para expresar la nacionalidad del receptor del comprobante que ampara retenciones e información de pagos.
        :param nacional: Nodo requerido para expresar la información del contribuyente receptor en caso de que sea de nacionalidad mexicana.
        :param extranjero: Nodo requerido para expresar la información del contribuyente receptor del comprobante que ampara retenciones e información de pagos, cuando sea residente en el extranjero.
        """

        super().__init__({
            'NacionalidadR': nacionalidad_r,
            'Nacional': nacional,
            'Extranjero': extranjero,
        })


class Emisor(ScalarMap):
    """
    Nodo requerido para expresar la información del contribuyente emisor del comprobante que ampara retenciones e información de pagos.

    :param rfc_e: Atributo requerido para registrar la clave del Registro Federal de Contribuyentes correspondiente al contribuyente emisor del comprobante que ampara retenciones e información de pagos, sin guiones o espacios.
    :param nom_den_raz_soc_e: Atributo requerido para registrar el nombre, denominación o razón social del contribuyente inscrito en el RFC, emisor del comprobante que ampara retenciones e información de pagos.
    :param regimen_fiscal_e: Atributo requerido para incorporar la clave del régimen del contribuyente emisor del comprobante que ampara retenciones e información de pagos.
    """

    def __init__(
            self,
            rfc_e: str,
            nom_den_raz_soc_e: str,
            regimen_fiscal_e: str,
    ):
        super().__init__({
            'RfcE': rfc_e,
            'NomDenRazSocE': nom_den_raz_soc_e,
            'RegimenFiscalE': regimen_fiscal_e,
        })


class Retenciones(CFDI):
    """
    Estándar del Comprobante Fiscal Digital por Internet que ampara retenciones e información de pagos. Los importes se expresan en la moneda de pesos mexicanos (MXN).
    """
    tag = '{http://www.sat.gob.mx/esquemas/retencionpago/2}Retenciones'
    version = '2.0'

    def __init__(
            self,
            lugar_exp_retenc: str,
            cve_retenc: str,
            emisor: Emisor | dict,
            receptor: Receptor | dict,
            periodo: Periodo | dict,
            totales: Totales | dict,
            folio_int: str = None,
            desc_retenc: str = None,
            cfdi_reten_relacionados: CfdiRetenRelacionados | dict = None,
            complemento: XElement | Sequence[XElement] = None,
            addenda: XElement | Sequence[XElement] = None,
            fecha_exp: datetime = None,
    ):
        """
        Estándar del Comprobante Fiscal Digital por Internet que ampara retenciones e información de pagos. Los importes se expresan en la moneda de pesos mexicanos (MXN).

        :param sello: Atributo requerido para contener el sello digital del comprobante que ampara retenciones e información de pagos, al que hacen referencia las reglas de resolución miscelánea vigente. El sello debe ser expresado como una cadena de texto en formato Base 64.
        :param no_certificado: Atributo requerido para expresar el número de serie del certificado de sello digital que ampara al comprobante de retención e información de pagos, de acuerdo con el acuse correspondiente a 20 posiciones otorgado por el sistema del SAT.
        :param certificado: Atributo requerido que sirve para incorporar el certificado de sello digital que ampara el comprobante de retención e información de pagos. El certificado debe ser expresado como una cadena de texto en formato Base 64.
        :param fecha_exp: Atributo requerido para la expresión de la fecha y hora de expedición del comprobante que ampara retenciones e información de pagos. Se expresa en la forma AAAA-MM-DDThh:mm:ss y debe corresponder con la hora local donde se expide el comprobante.
        :param lugar_exp_retenc: Atributo requerido para incorporar el código postal del lugar de expedición del comprobante que ampara retenciones e información de pagos.
        :param cve_retenc: Atributo requerido para expresar la clave de la retención e información de pagos de acuerdo al catálogo publicado en internet por el SAT.
        :param emisor: Nodo requerido para expresar la información del contribuyente emisor del comprobante que ampara retenciones e información de pagos.
        :param receptor: Nodo requerido para expresar la información del contribuyente receptor del comprobante que ampara retenciones e información de pagos.
        :param periodo: Nodo requerido para expresar el período que corresponde al comprobante que ampara retenciones e información de pagos.
        :param totales: Nodo requerido para expresar el total de las retenciones e información de los pagos efectuados en el período que ampara el comprobante.
        :param folio_int: Atributo opcional para control interno del contribuyente que expresa el folio del comprobante que ampara retenciones e información de pagos. Permite números y/o letras.
        :param desc_retenc: Atributo condicional que expresa la descripción de la retención e información de pagos.
        :param cfdi_reten_relacionados: Nodo opcional para precisar la información de los comprobantes relacionados.
        :param complemento: Nodo opcional donde se incluirá el complemento Timbre Fiscal Digital de manera obligatoria y los nodos complementarios determinados por el SAT, de acuerdo a las disposiciones particulares a un sector o actividad específica.
        :param addenda: Nodo opcional para recibir las extensiones al formato que sean de utilidad al contribuyente. Para las reglas de uso del mismo, referirse al formato de origen.
        """

        fecha_exp = fecha_exp or datetime.now(tz=get_timezone(lugar_exp_retenc)).replace(tzinfo=None)
        super().__init__({
            'Version': self.version,
            'Sello': '',
            'FechaExp': fecha_exp,
            'LugarExpRetenc': lugar_exp_retenc,
            'CveRetenc': cve_retenc,
            'FolioInt': folio_int,
            'DescRetenc': desc_retenc,
            'CfdiRetenRelacionados': cfdi_reten_relacionados,
            'Emisor': emisor,
            'Receptor': receptor,
            'Periodo': periodo,
            'Totales': totales,
            'Complemento': complemento,
            'Addenda': addenda,
        })

    def sign(self, signer: Signer):
        self['NoCertificado'] = signer.certificate_number,
        self['Certificado'] = signer.certificate_base64(),
        self['Sello'] = signer.sign_sha256(
            self.cadena_original().encode()
        )



import os
from datetime import datetime, date
from decimal import Decimal
from unittest import mock

import pytest

from satcfdi.cfdi import CFDI
from satcfdi.create.cfd import cfdi33, nomina12
from satcfdi.create.cfd.cfdi40 import PagoComprobante
from satcfdi.pacs.sat import SAT
from tests.utils import get_signer, verify_result, _uuid, stamp_v11, SAT_Certificate_Store_Pruebas, XElementPrettyPrinter

module = 'satcfdi'
current_dir = os.path.dirname(__file__)
current_filename = os.path.splitext(os.path.basename(__file__))[0]
sat = SAT()

invoices = [
    ('xiqb891116qe4', "xiqb891116qe4_ingreso_noobjeto", None, [cfdi33.Impuesto.parse('001|Tasa|0.100000')], '13851.27', False),
    # ('xiqb891116qe4', "xiqb891116qe4_ingreso_exento", cfdi33.Traslado.IVA_EXENTO, [cfdi33.Retencion.ISR10], '13851.27', False),
    ('xiqb891116qe4', "xiqb891116qe4_ingreso_iva16", cfdi33.Impuesto.parse('002|Tasa|0.160000'), [cfdi33.Impuesto.parse('001|Tasa|0.100000'), cfdi33.Impuesto.parse('002|Tasa|0.106667')], '14672.08', False),
    ('h&e951128469', "h&e951128469_ingreso_noobjeto", None, None, '15390.30', False),
    # ('h&e951128469', "h&e951128469_ingreso_exento", cfdi33.Traslado.IVA_EXENTO, None, '15390.30', False),
    ('h&e951128469', "h&e951128469_ingreso_iva16", cfdi33.Impuesto.parse('002|Tasa|0.160000'), None, '17852.75', False),
    # ('h&e951128469', "h&e951128469_ingreso_ieps_exento", cfdi33.Traslado.IEPS_EXENTO, None, '15390.30', False)
]


def verify_invoice(invoice, path, include_metadata=False):
    if include_metadata:
        verify = verify_result(data=invoice.metadata, filename=f"{path}.txt")
        assert verify

    pp = XElementPrettyPrinter()
    verify = verify_result(data=pp.pformat(invoice), filename=f"{path}.pretty.py")
    assert verify

    verify = verify_result(data=invoice.xml_bytes(pretty_print=True), filename=f"{path}.xml")
    assert verify

    verify = verify_result(data=invoice.html_str(), filename=f"{path}.html")
    assert verify


@pytest.mark.parametrize('rfc, xml_file, traslados, retenciones, total, traslado_incluido', invoices)
def test_create_invoice(rfc, xml_file, traslados, retenciones, total, traslado_incluido):
    signer = get_signer(rfc)
    emisor = cfdi33.Emisor(
        rfc=signer.rfc,
        nombre=signer.legal_name,
        regimen_fiscal="601"
    )

    invoice = cfdi33.Comprobante(
        emisor=emisor,
        lugar_expedicion="56820",
        fecha=datetime.fromisoformat("2020-01-01T22:40:38"),
        receptor=cfdi33.Receptor(
            rfc='KIJ0906199R1',
            nombre='KIJ, S.A DE C.V.',
            uso_cfdi='G03'
        ),
        metodo_pago='PPD',
        serie="A",
        folio="123456",
        conceptos=[
            cfdi33.Concepto(
                cuenta_predial='1234567890',
                clave_prod_serv='10101702',
                cantidad=Decimal('1.00'),
                clave_unidad='E48',
                descripcion='SERVICIOS DE FACTURACION',
                valor_unitario=Decimal('15390.30'),
                impuestos=cfdi33.Impuestos(
                    traslados=traslados,
                    retenciones=retenciones
                )
            )
        ]
    )
    invoice.sign(signer)
    invoice.to_xml(validate=True)
    assert invoice["Total"] == Decimal(total)

    verify_invoice(invoice, f"{xml_file}")

    # Stamping
    with mock.patch('uuid.uuid4', _uuid), \
            mock.patch(f'{module}.Certificate.rfc_pac', "SAT970701NN3"):
        stamp_v11(invoice, signer=signer, date=datetime(year=2020, month=1, day=11))

    verify_invoice(invoice, f"{xml_file}_stamped")

    def from_no_certificado(_, no_certificado):
        return signer

    # Verify Signature
    with mock.patch(f'{module}.transform.SAT_Certificate_Store', SAT_Certificate_Store_Pruebas), \
            mock.patch(f'{module}.pacs.sat.SAT.recover_certificate', from_no_certificado), \
            mock.patch(f'{module}.Certificate.rfc_pac', "SAT970701NN3"):
        ver = sat.validate(invoice)
        assert ver


@pytest.mark.parametrize('rfc, xml_file, traslados, retenciones, total, traslado_incluido', invoices)
def test_create_pago(rfc, xml_file, traslados, retenciones, total, traslado_incluido):
    signer = get_signer(rfc)
    emisor = cfdi33.Emisor(
        rfc=signer.rfc,
        nombre=signer.legal_name,
        regimen_fiscal="601"
    )

    ingreso_invoice = CFDI.from_file(os.path.join(current_dir, f"{current_filename}/{xml_file}_stamped.xml"))

    invoice = cfdi33.Comprobante.pago_comprobantes(
        emisor=emisor,
        lugar_expedicion="56820",
        fecha=datetime.fromisoformat("2020-01-01T22:40:38"),
        comprobantes=[ingreso_invoice],
        fecha_pago=datetime.fromisoformat("2020-01-02T22:40:38"),
        forma_pago="03",
        serie="A",
        folio="123456",
    )
    invoice.sign(signer)

    assert invoice['Complemento']['Pago'][0]['Monto'] == Decimal(total)

    verify_invoice(invoice, f"pago_{xml_file}")


@pytest.mark.parametrize('rfc, xml_file, traslados, retenciones, total, traslado_incluido', invoices)
def test_create_pago_parcial(rfc, xml_file, traslados, retenciones, total, traslado_incluido):
    signer = get_signer(rfc)
    emisor = cfdi33.Emisor(
        rfc=signer.rfc,
        nombre=signer.legal_name,
        regimen_fiscal="601"
    )

    ingreso_invoice = CFDI.from_file(os.path.join(current_dir, f"{current_filename}/{xml_file}_stamped.xml"))

    invoice = cfdi33.Comprobante.pago_comprobantes(
        emisor=emisor,
        lugar_expedicion="56820",
        fecha=datetime.fromisoformat("2020-01-01T22:40:38"),
        comprobantes=PagoComprobante(
            comprobante=ingreso_invoice,
            num_parcialidad=2,
            imp_saldo_ant=Decimal("5000.43"),
            imp_pagado=Decimal("3245.12"),
        ),
        fecha_pago=datetime.fromisoformat("2020-01-02T22:40:38"),
        forma_pago="03",
        serie="A",
        folio="123456",
    )
    invoice.sign(signer)

    assert invoice['Complemento']['Pago'][0]['Monto'] == Decimal("3245.12")

    verify_invoice(invoice, f"pago_p_{xml_file}")


@pytest.mark.parametrize('rfc, xml_file, traslados, retenciones, total, traslado_incluido', invoices)
def test_create_pago_multiple(rfc, xml_file, traslados, retenciones, total, traslado_incluido):
    signer = get_signer(rfc)
    emisor = cfdi33.Emisor(
        rfc=signer.rfc,
        nombre=signer.legal_name,
        regimen_fiscal="601"
    )

    ingreso_invoice = CFDI.from_file(os.path.join(current_dir, f"{current_filename}/{xml_file}_stamped.xml"))

    invoice = cfdi33.Comprobante.pago_comprobantes(
        emisor=emisor,
        lugar_expedicion="56820",
        fecha=datetime.fromisoformat("2020-01-01T22:40:38"),
        comprobantes=[ingreso_invoice, ingreso_invoice],
        fecha_pago=datetime.fromisoformat("2020-01-02T22:40:38"),
        forma_pago="03",
        serie="A",
        folio="123456",
    )
    invoice.sign(signer)

    assert invoice['Complemento']['Pago'][0]['Monto'] == Decimal(total) * 2

    verify_invoice(invoice, f"mpago_{xml_file}")


def test_nomina():
    xml_file = "invoice_nomina"

    signer = get_signer('xiqb891116qe4')
    emisor = cfdi33.Emisor(
        rfc=signer.rfc,
        nombre=signer.legal_name,
        regimen_fiscal="601"
    )

    invoice = cfdi33.Comprobante.nomina(
        emisor=emisor,
        receptor=cfdi33.Receptor(
            rfc='KIJ0906199R1',
            nombre='KIJ, S.A DE C.V.',
            uso_cfdi='G03',
        ),
        lugar_expedicion="56820",
        complemento_nomina=nomina12.Nomina(
            emisor={
                'RegistroPatronal': 'Z1234567890'
            },
            receptor={
                'Curp': 'XIQB891116MCHZRL72',
                'NumSeguridadSocial': '987654321',
                'FechaInicioRelLaboral': date(2020, 1, 1),
                'Antigüedad': 'P96W',
                'TipoContrato': '01',
                'Sindicalizado': 'No',
                'TipoJornada': '01',
                'TipoRegimen': '02',
                'NumEmpleado': '12345678',
                'Departamento': 'Departamento del Trabajo',
                'Puesto': 'Trabajador',
                'RiesgoPuesto': '2',
                'PeriodicidadPago': '04',
                'CuentaBancaria': '0001000200030004',
                'SalarioDiarioIntegrado': Decimal('100.01'),
                'ClaveEntFed': 'MOR'
            },
            percepciones={
                'Percepcion': [
                    {
                        'TipoPercepcion': '001',
                        'Clave': '001',
                        'Concepto': 'SUELDO',
                        'ImporteGravado': Decimal('1200'),
                        'ImporteExento': Decimal('400')
                    }
                ]
            },
            deducciones={
                'Deduccion': [
                    {
                        'TipoDeduccion': '002',
                        'Clave': '300',
                        'Concepto': 'ISR A CARGO',
                        'Importe': Decimal('1234.73')
                    },
                    {
                        'TipoDeduccion': '004',
                        'Clave': '204',
                        'Concepto': 'FONDO DE AHORRO EMPLEADOS',
                        'Importe': Decimal('521.10')
                    },
                ]
            },
            otros_pagos=[
                {
                    'TipoOtroPago': '999',
                    'Clave': '046',
                    'Concepto': 'REEMBOLSO DE GASTOS',
                    'Importe': Decimal('456')
                },
                {
                    'SubsidioAlEmpleo': Decimal('0'),
                    'TipoOtroPago': '002',
                    'Clave': '002',
                    'Concepto': 'SUBSIDIO EMPLEO',
                    'Importe': Decimal('0')
                }
            ],
            incapacidades=[
                {
                    "DiasIncapacidad": 2,
                    'TipoIncapacidad': '02',
                    'ImporteMonetario': Decimal(1000),
                }
            ],
            tipo_nomina='O',
            fecha_pago=date(2020, 1, 30),
            fecha_final_pago=date(2020, 1, 31),
            fecha_inicial_pago=date(2020, 1, 16),
            num_dias_pagados=Decimal('16.000')
        ),
        serie="A",
        folio="123456",
        fecha=datetime.fromisoformat("2020-09-29T22:40:38")
    )
    invoice.sign(signer)

    verify_invoice(invoice, f"{xml_file}")



"""AuxiliarCtas www.sat.gob.mx/esquemas/ContabilidadE/1_1/AuxiliarCtas"""
from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ... import CFDI, XElement, ScalarMap


class DetalleAux(ScalarMap):
    """
    Nodo obligatorio para expresar el detalle de los movimientos del periodo de cada uno de los auxiliares
    
    :param fecha: Atributo requerido para expresar la fecha de registro de la transacción que afecta la cuenta o subcuenta que integra el auxiliar.
    :param num_un_iden_pol: Atributo requerido para expresar el número único de identificación de la póliza. El campo deberá contener la clave o nombre utilizado por el contribuyente para diferenciar, el tipo de póliza y el número correspondiente. En un mes ordinario no debe repetirse un mismo número de póliza con la clave o nombre asignado por el contribuyente.
    :param concepto: Atributo requerido para expresar el concepto de la transacción que integra el auxiliar.
    :param debe: Atributo requerido para expresar el monto del cargo de la cuenta o subcuenta de la transacción que integra el auxiliar. En caso de no existir dato, colocar cero (0)
    :param haber: Atributo requerido para expresar el monto del abono de la cuenta o subcuenta de la transacción que integra el auxiliar. En caso de no existir dato, colocar cero (0)
    """
    
    def __init__(
            self,
            fecha: date,
            num_un_iden_pol: str,
            concepto: str,
            debe: Decimal | int,
            haber: Decimal | int,
    ): 
        super().__init__({
            'Fecha': fecha,
            'NumUnIdenPol': num_un_iden_pol,
            'Concepto': concepto,
            'Debe': debe,
            'Haber': haber,
        })
        

class Cuenta(ScalarMap):
    """
    Nodo obligatorio para expresar los movimientos del periodo de cada uno de los auxiliares de la cuenta y/o subcuenta.
    
    :param num_cta: Atributo requerido para expresar la clave con que se distingue la cuenta o subcuenta que se afecta por la transacción que integra el auxiliar.
    :param des_cta: Atributo requerido para expresar el concepto de la cuenta o subcuenta que se afecta por la transacción que integra el auxiliar.
    :param saldo_ini: Atributo requerido para expresar el monto del saldo inicial de la cuenta o subcuenta del periodo del auxiliar. En caso de no existir dato, colocar cero (0)
    :param saldo_fin: Atributo requerido para expresar el monto del saldo final de la cuenta o subcuenta del periodo del auxiliar. En caso de no existir dato, colocar cero (0)
    :param detalle_aux: Nodo obligatorio para expresar el detalle de los movimientos del periodo de cada uno de los auxiliares
    """
    
    def __init__(
            self,
            num_cta: str,
            des_cta: str,
            saldo_ini: Decimal | int,
            saldo_fin: Decimal | int,
            detalle_aux: DetalleAux | dict | Sequence[DetalleAux | dict],
    ): 
        super().__init__({
            'NumCta': num_cta,
            'DesCta': des_cta,
            'SaldoIni': saldo_ini,
            'SaldoFin': saldo_fin,
            'DetalleAux': detalle_aux,
        })
        

class AuxiliarCtas(CFDI):
    """
    Estándar de auxiliar de la cuenta o subcuenta del periodo que se entrega como parte de la contabilidad electrónica
    
    :param rfc: Atributo requerido para expresar el RFC del contribuyente que envía los datos.
    :param mes: Atributo requerido para expresar el mes en que inicia la vigencia del auxiliar de la cuenta o subcuenta.
    :param anio: Atributo requerido para expresar el año al que inicia la vigencia del auxiliar de la cuenta o subcuenta.
    :param tipo_solicitud: Atributo requerido para expresar el tipo de envío del auxiliar de la cuenta o subcuenta ( AF - Acto de Fiscalización; FC - Fiscalización Compulsa; DE - Devolución; CO - Compensación )
    :param cuenta: Nodo obligatorio para expresar los movimientos del periodo de cada uno de los auxiliares de la cuenta y/o subcuenta.
    :param num_orden: Atributo opcional para expresar el número de orden asignado al acto de fiscalización al que hace referencia la solicitud del auxiliar de la cuenta o subcuenta. Requerido para tipo de solicitud = AF y FC. Se convierte en requerido cuando se cuente con la información.
    :param num_tramite: Atributo opcional para expresar el número de trámite asignado a la solicitud de devolución o compensación al que hace referencia el auxiliar de la cuenta o subcuenta. Requerido para tipo de solicitud = DE o CO. Se convierte en requerido cuando se cuente con la información.
    :param sello: Atributo opcional para contener el sello digital del archivo de contabilidad electrónica. El sello deberá ser expresado cómo una cadena de texto en formato Base 64
    :param no_certificado: Atributo opcional para expresar el número de serie del certificado de sello digital que ampara el archivo de contabilidad electrónica, de acuerdo al acuse correspondiente a 20 posiciones otorgado por el sistema del SAT.
    :param certificado: Atributo opcional que sirve para expresar el certificado de sello digital que ampara al archivo de contabilidad electrónica como texto, en formato base 64.
    """
    
    tag = '{www.sat.gob.mx/esquemas/ContabilidadE/1_1/AuxiliarCtas}AuxiliarCtas'
    version = '1.1'
    
    def __init__(
            self,
            rfc: str,
            mes: str,
            anio: int,
            tipo_solicitud: str,
            cuenta: Cuenta | dict | Sequence[Cuenta | dict],
            num_orden: str = None,
            num_tramite: str = None,
            sello: str = None,
            no_certificado: str = None,
            certificado: str = None,
    ): 
        super().__init__({
            'Version': self.version,
            'RFC': rfc,
            'Mes': mes,
            'Anio': anio,
            'TipoSolicitud': tipo_solicitud,
            'Cuenta': cuenta,
            'NumOrden': num_orden,
            'NumTramite': num_tramite,
            'Sello': sello,
            'NoCertificado': no_certificado,
            'Certificado': certificado,
        })
        




"""premios http://www.sat.gob.mx/esquemas/retencionpago/1/premios"""
from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ... import CFDI, XElement, ScalarMap


class Premios(CFDI):
    """
    Complemento para expresar los premios obtenidos en un periodo o ejercicio
    
    :param entidad_federativa: Atributo requerido para expresar la entidad federativa en la que se paga el premio obtenido, conforme al catálogo.
    :param mont_tot_pago: Atributo requerido para expresar el importe del pago realizado por la obtención de un premio
    :param mont_tot_pago_grav: Atributo requerido para expresar el importe gravado en la obtención de un premio
    :param mont_tot_pago_exent: Atributo requerido para expresar el monto total exento en la obtención de un premio
    """
    
    tag = '{http://www.sat.gob.mx/esquemas/retencionpago/1/premios}Premios'
    version = '1.0'
    
    def __init__(
            self,
            entidad_federativa: str,
            mont_tot_pago: Decimal | int,
            mont_tot_pago_grav: Decimal | int,
            mont_tot_pago_exent: Decimal | int,
    ): 
        super().__init__({
            'Version': self.version,
            'EntidadFederativa': entidad_federativa,
            'MontTotPago': mont_tot_pago,
            'MontTotPagoGrav': mont_tot_pago_grav,
            'MontTotPagoExent': mont_tot_pago_exent,
        })
        




from datetime import datetime
import io
import os
from unittest import mock
from zipfile import ZipInfo

import pytest
from cryptography.hazmat.primitives.serialization import load_der_private_key

from satcfdi.certifica import _create_certificate_signing_request, _create_certificate_signing_request_zip, Certifica, _calculate_code, \
    _create_renovation_certificate_signing_request, _calculate_code_random, _create_generacion_certificate_signing_request, _create_renovation_moral_certificate_signing_request
from tests.utils import get_signer

current_dir = os.path.dirname(__file__)
current_filename = os.path.splitext(os.path.basename(__file__))[0]
test_dir = os.path.join(current_dir, current_filename)
module = 'satcfdi'


def test_certificate_csd():
    signer = get_signer("cacx7605101p8")
    sucursal = 'TestSucursal'

    with mock.patch(f'{module}.certifica.datetime') as d:
        d.now = mock.Mock(return_value=datetime(2022, 11, 18, 11, 7, 10))

        certifica = Certifica(signer=signer)
        certifica.solicitud_certificado(
            sucursal=sucursal,
            password="TestIt1234",
            dirname=os.path.join(test_dir, 'test_csd')
        )

    private_key = load_der_private_key(
        data=open(os.path.join(test_dir, 'test_csd', 'CSD_TestSucursal_CACX7605101P8_20221118_110710.key'), 'rb').read(),
        password=b'TestIt1234'
    )

    print(private_key)


def test_calculate_code():
    signer = get_signer("HAÑ930228SM9")
    assert _calculate_code(signer) == b'+HcCMz3yWNA/QxB/l2flfHZ1slU='
    signer = get_signer("cacx7605101p8")
    assert _calculate_code(signer) == b'g3dhTGuMM+KZoCQ37nuuG9IUzKk='
    signer = get_signer("IAÑL750210963")
    assert len(_calculate_code(signer)) == len(_calculate_code_random())


generacion = [
    ('generacion_1', 'Test1234', "IAÑL750210963", datetime(2022, 11, 20, 0, 44, 33), 'micorreo@pruebas.gob.mx',
     b'T/pD4GwyUGB9QOzYpWKLikLuP6U=', 'IANL750210HSRNZS09', "IAÑL750210963"),
    ('generacion_2', 'Test1234', "HAÑ930228SM9", datetime(2022, 11, 20, 21, 18, 37), 'super@correo.com',
     b'gyUkJAqQytfSsWrQSeC/ZOvgz6o=', ' / ', 'HAÑ930228SM9 / IAÑL750210963'),
]


@pytest.mark.parametrize('folder, password, rfc, file_name_date, correo, code, curp, rfc_rep', generacion)
def test_generacion_request(folder, password, rfc, file_name_date, correo, code, curp, rfc_rep):
    tmf = file_name_date.strftime('%Y%m%d_%H%M%S')

    private_key = load_der_private_key(
        data=open(os.path.join(test_dir, folder, f'Claveprivada_FIEL_{rfc}_{tmf}.key'), 'rb').read(),
        password=password.encode('windows-1252')
    )

    assert private_key.key_size == 2048
    assert private_key.public_key().public_numbers().e == 65537

    with mock.patch(f"{module}.certifica._calculate_code_random", lambda: code):
        res = _create_generacion_certificate_signing_request(private_key=private_key, rfc=rfc_rep, curp=curp, correo=correo)

    cer_req = open(os.path.join(test_dir, folder, f"Requerimiento_FIEL_{rfc}_{tmf}.req"), 'rb').read()
    assert res == cer_req


renovaciones = [
    ('renovacion_1', 'Test_iañl750210963!', "IAÑL750210963", datetime(2022, 11, 19, 21, 2, 18), 'micorreo@pruebas.gob.mx',
     b'vNJ4zp78Y7z75+lq99ltUJQPkjw='),
]


@pytest.mark.parametrize('folder, password, rfc, file_name_date, correo, code', renovaciones)
def test_create_renovacion_request(folder, password, rfc, file_name_date, correo, code):
    signer = get_signer(rfc)
    tmf = file_name_date.strftime('%Y%m%d_%H%M%S')

    private_key = load_der_private_key(
        data=open(os.path.join(test_dir, folder, f'Claveprivada_FIEL_{rfc}_{tmf}.key'), 'rb').read(),
        password=password.encode('windows-1252')
    )

    assert private_key.key_size == 2048
    assert private_key.public_key().public_numbers().e == 65537

    with mock.patch(f"{module}.certifica._calculate_code_random", lambda: code):
        res = _create_renovation_certificate_signing_request(signer=signer, private_key=private_key, correo=correo)

    cer_req = open(os.path.join(test_dir, folder, f"Renovacion_FIEL_{rfc}_{tmf}.req"), 'rb').read()
    assert res == cer_req


renovaciones_moral = [
    ('renovacion_moral_1', 'Test1234', "HAÑ930228SM9", "KAHO641101B39", datetime(2022, 11, 20, 22, 16, 25), 'pruebas.sat@pruebas.gob.mx',
     b'BMwmgQnOpww0B3x7EhvlPsrywFw='),
]


@pytest.mark.parametrize('folder, password, rfc, rfc_rep, file_name_date, correo, code', renovaciones_moral)
def test_create_renovacion_moral_request(folder, password, rfc, rfc_rep, file_name_date, correo, code):
    signer = get_signer(rfc_rep)
    tmf = file_name_date.strftime('%Y%m%d_%H%M%S')

    private_key = load_der_private_key(
        data=open(os.path.join(test_dir, folder, f'Claveprivada_FIEL_{rfc}_{tmf}.key'), 'rb').read(),
        password=password.encode('windows-1252')
    )

    assert private_key.key_size == 2048
    assert private_key.public_key().public_numbers().e == 65537

    with mock.patch(f"{module}.certifica._calculate_code_random", lambda: code):
        res = _create_renovation_moral_certificate_signing_request(signer=signer, private_key=private_key, rfc=rfc, correo=correo)

    cer_req = open(os.path.join(test_dir, folder, f"Renovacion_FIEL_{rfc}_{tmf}.req"), 'rb').read()
    assert res == cer_req


solicitudes = [
    ('solicitud_1', 'Test1234', (2022, 11, 18, 11, 7, 22), 'MiSucursal', "CACX7605101P8", datetime(2022, 11, 18, 11, 7, 10)),
    ('solicitud_2', 'Test&ñ200!', (2022, 11, 18, 15, 32, 44), 'ABC& hañ 200!', "HAÑ930228SM9", datetime(2022, 11, 18, 15, 32, 25)),
]


@pytest.mark.parametrize('folder, password, zip_date, sucursal, rfc, file_name_date', solicitudes)
def test_create_signing_request(folder, password, zip_date, sucursal, rfc, file_name_date):
    signer = get_signer(rfc)
    tmf = file_name_date.strftime('%Y%m%d_%H%M%S')
    sucursal_f = sucursal.replace(" ", "_")

    private_key = load_der_private_key(
        data=open(os.path.join(test_dir, folder, f'CSD_{sucursal_f}_{rfc}_{tmf}.key'), 'rb').read(),
        password=password.encode('windows-1252')
    )

    assert private_key.key_size == 2048
    assert private_key.public_key().public_numbers().e == 65537

    res = _create_certificate_signing_request(signer=signer, private_key=private_key, sucursal=sucursal)

    cer_req = open(os.path.join(test_dir, folder, f"CSD_{sucursal_f}_{rfc}_{tmf}s.req"), 'rb').read()
    assert res == cer_req


@pytest.mark.parametrize('folder, password, zip_date, sucursal, rfc, file_name_date', solicitudes)
def test_create_signing_request_zip(folder, password, zip_date, sucursal, rfc, file_name_date):
    signer = get_signer(rfc)
    tmf = file_name_date.strftime('%Y%m%d_%H%M%S')
    sucursal_f = sucursal.replace(" ", "_")

    private_key = load_der_private_key(
        data=open(os.path.join(test_dir, folder, f'CSD_{sucursal_f}_{rfc}_{tmf}.key'), 'rb').read(),
        password=password.encode('windows-1252')
    )

    # Verify ZIP
    def zi(filename):
        return ZipInfo(
            filename=filename,
            date_time=zip_date  # time.localtime(time.time())[:6]
        )

    with mock.patch(f'{module}.zip.ZipInfo', zi) as m, mock.patch(f'{module}.certifica.datetime') as d:
        d.now = mock.Mock(return_value=file_name_date)

        res = _create_certificate_signing_request_zip(signer=signer, private_key=private_key, sucursal=sucursal)

    zip = open(os.path.join(test_dir, folder, f"CSD_{sucursal_f}_{rfc}_{tmf}s.zip"), 'rb').read()
    assert res == zip

    certifica = Certifica(signer)
    res = certifica._pkcs7_package(zip)

    assert len(res) > 3200
    # check something?



"""spei http://www.sat.gob.mx/spei"""
from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ... import CFDI, XElement, ScalarMap


class Beneficiario(ScalarMap):
    """
    Elemento para describir los datos del beneficiario del SPEI
    
    :param banco_receptor: Atributo requerido para expresar el nombre del Banco o Institución Financiera Receptora del SPEI
    :param nombre: Nombre de la persona física o moral receptora del pago.
    :param tipo_cuenta: Categoría de la cuenta a la que se efectuará el abono por la transferencia electrónica de fondos. Consultar Catálogo de Tipos de Cuenta.
    :param cuenta: Esta cuenta deberá estar ligada al campo Tipo de Cuenta del Beneficiario, donde son abonados los fondos.
    :param rfc: Atributo requerido para la expresión del registro federal de contribuyentes del beneficiario. Se pondrá ND en caso de no estar disponible
    :param concepto: Descripción del motivo por el que el ordenante hace el pago al beneficiario.
    :param monto_pago: Atributo obligatorio para la expresión del monto de la operación. Se trata de un entero positivo
    :param iva: Importes de IVA correspondientes al pago. El monto debe ser mayor a cero y menor o igual a 9,999,999,999,999,999.99
    """
    
    def __init__(
            self,
            banco_receptor: str,
            nombre: str,
            tipo_cuenta: Decimal | int,
            cuenta: Decimal | int,
            rfc: str,
            concepto: str,
            monto_pago: Decimal | int,
            iva: Decimal | int = None,
    ): 
        super().__init__({
            'BancoReceptor': banco_receptor,
            'Nombre': nombre,
            'TipoCuenta': tipo_cuenta,
            'Cuenta': cuenta,
            'RFC': rfc,
            'Concepto': concepto,
            'MontoPago': monto_pago,
            'IVA': iva,
        })
        

class Ordenante(ScalarMap):
    """
    Elemento para describir los datos del ordenante del SPEI
    
    :param banco_emisor: Atributo requerido para expresar el nombre del Banco o Institución Financiera emisora del SPEI
    :param nombre: Nombre de la persona física o moral que ordena el envío del pago.
    :param tipo_cuenta: Categoría de la Cuenta a la que se efectuará el cargo por la transferencia electrónica de fondos.
    :param cuenta: Cuenta que deberá estar ligada al Tipo de Cuenta del Ordenante, donde serán cargados los fondos.
    :param rfc: Corresponde al registro federal de contribuyentes o clave única de registro de población del ordenante. Se pondrá ND en caso de no tenerlo disponible
    """
    
    def __init__(
            self,
            banco_emisor: str,
            nombre: str,
            tipo_cuenta: Decimal | int,
            cuenta: Decimal | int,
            rfc: str,
    ): 
        super().__init__({
            'BancoEmisor': banco_emisor,
            'Nombre': nombre,
            'TipoCuenta': tipo_cuenta,
            'Cuenta': cuenta,
            'RFC': rfc,
        })
        

class SPEI_Tercero(ScalarMap):
    """
    Estándar aplicable a operaciones de SPEI a terceros
    
    :param fecha_operacion: Fecha de operación con formato. Debe ser la misma que la fecha de operación del sistema.
    :param hora: hora del acreditamiento
    :param clave_spei: Clave SPEI del Participante Emisor.
    :param sello: Atributo requerido para contener el sello digital del comprobante de pago. El sello deberá ser expresado cómo una cadena de texto en formato Base 64.
    :param numero_certificado: Atributo requerido para la identificación del certificado de seguridad utilizado para el sello digital.
    :param cadena_cda: Atributo que contiene la información del CDA fidedigna que la institución ha enviado a Banco de México.
    :param ordenante: Elemento para describir los datos del ordenante del SPEI
    :param beneficiario: Elemento para describir los datos del beneficiario del SPEI
    """
    
    def __init__(
            self,
            fecha_operacion: date,
            hora: time,
            clave_spei: int,
            sello: str,
            numero_certificado: str,
            cadena_cda: str,
            ordenante: Ordenante | dict,
            beneficiario: Beneficiario | dict,
    ): 
        super().__init__({
            'FechaOperacion': fecha_operacion,
            'Hora': hora,
            'ClaveSPEI': clave_spei,
            'Sello': sello,
            'NumeroCertificado': numero_certificado,
            'CadenaCDA': cadena_cda,
            'Ordenante': ordenante,
            'Beneficiario': beneficiario,
        })
        

class ComplementoSPEI(CFDI):
    """
    Complemento para el uso de SPEI Tercero a Tercero
    
    :param spei_tercero: Estándar aplicable a operaciones de SPEI a terceros
    """
    
    tag = '{http://www.sat.gob.mx/spei}Complemento_SPEI'
    
    def __init__(
            self,
            spei_tercero: SPEI_Tercero | dict | Sequence[SPEI_Tercero | dict],
    ): 
        super().__init__({
            'SPEI_Tercero': spei_tercero,
        })
        




"""psgecfd http://www.sat.gob.mx/psgecfd"""
from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ... import CFDI, XElement, ScalarMap


class PrestadoresDeServiciosDeCFD(CFDI):
    """
    Complemento requerido para los Proveedores de Servicio de generación y envío de Comprobantes Fiscales Digitales.
    
    :param nombre: Atributo requerido para el nombre o razón social del proveedor de servicios autorizado que generó y envió el comprobante fiscal digital.
    :param rfc: Atributo requerido para el Registro Federal de Contribuyentes del proveedor de servicios de generación y envío de comprobantes fiscales digitales.
    :param no_certificado: Atributo requerido para expresar el número de serie del certificado de sello digital del proveedor del servicio autorizado que generó y envió el comprobante fiscal digital.
    :param fecha_autorizacion: Atributo requerido para la expresión de la fecha y hora de autorización del proveedor del servicio que generó y envió el comprobante fiscal digital. Se expresa en la forma aaaa-mm-ddThh:mm:ss, de acuerdo con la especificación ISO 8601.
    :param no_autorizacion: Atributo requerido para precisar el número de autorización del proveedor del servicio de generación y envío comprobantes autorizado por el SAT.
    :param sello_del_psgecfd: Atributo requerido para contener el sello digital del proveedor del servicio de generación y envío de comprobantes fiscales digitales que generó y emitió el comprobante fiscal digital. El sello deberá ser expresado cómo una cadena de texto en formato Base 64.
    """
    
    tag = '{http://www.sat.gob.mx/psgecfd}PrestadoresDeServiciosDeCFD'
    
    def __init__(
            self,
            nombre: str,
            rfc: str,
            no_certificado: str,
            fecha_autorizacion: datetime,
            no_autorizacion: int,
            sello_del_psgecfd: str,
    ): 
        super().__init__({
            'Nombre': nombre,
            'Rfc': rfc,
            'NoCertificado': no_certificado,
            'FechaAutorizacion': fecha_autorizacion,
            'NoAutorizacion': no_autorizacion,
            'SelloDelPSGECFD': sello_del_psgecfd,
        })
        




"""mpc http://www.uif.shcp.gob.mx/recepcion/mpc"""
from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ... import CFDI, XElement, ScalarMap


class DatosLiquidacionType(ScalarMap):
    """
    
    :param fecha_disposicion:
    :param instrumento_monetario:
    :param moneda:
    :param monto_operacion:
    """
    
    def __init__(
            self,
            fecha_disposicion: date,
            instrumento_monetario: str,
            moneda: str,
            monto_operacion: str,
    ): 
        super().__init__({
            'FechaDisposicion': fecha_disposicion,
            'InstrumentoMonetario': instrumento_monetario,
            'Moneda': moneda,
            'MontoOperacion': monto_operacion,
        })
        

class FideicomisoGaranteType(ScalarMap):
    """
    
    :param denominacion_razon:
    :param rfc:
    :param identificador_fideicomiso:
    """
    
    def __init__(
            self,
            denominacion_razon: str,
            rfc: str = None,
            identificador_fideicomiso: str = None,
    ): 
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'Rfc': rfc,
            'IdentificadorFideicomiso': identificador_fideicomiso,
        })
        

class PersonaMoralGaranteType(ScalarMap):
    """
    
    :param denominacion_razon:
    :param fecha_constitucion:
    :param rfc:
    """
    
    def __init__(
            self,
            denominacion_razon: str,
            fecha_constitucion: date = None,
            rfc: str = None,
    ): 
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'FechaConstitucion': fecha_constitucion,
            'Rfc': rfc,
        })
        

class PersonaFisicaGaranteType(ScalarMap):
    """
    
    :param nombre:
    :param apellido_paterno:
    :param apellido_materno:
    :param fecha_nacimiento:
    :param rfc:
    :param curp:
    """
    
    def __init__(
            self,
            nombre: str,
            apellido_paterno: str,
            apellido_materno: str,
            fecha_nacimiento: date = None,
            rfc: str = None,
            curp: str = None,
    ): 
        super().__init__({
            'Nombre': nombre,
            'ApellidoPaterno': apellido_paterno,
            'ApellidoMaterno': apellido_materno,
            'FechaNacimiento': fecha_nacimiento,
            'Rfc': rfc,
            'Curp': curp,
        })
        

class TipoGaranteType(ScalarMap):
    """
    
    :param persona_fisica:
    :param persona_moral:
    :param fideicomiso:
    """
    
    def __init__(
            self,
            persona_fisica: PersonaFisicaGaranteType | dict = None,
            persona_moral: PersonaMoralGaranteType | dict = None,
            fideicomiso: FideicomisoGaranteType | dict = None,
    ): 
        super().__init__({
            'PersonaFisica': persona_fisica,
            'PersonaMoral': persona_moral,
            'Fideicomiso': fideicomiso,
        })
        

class DatosOtroType(ScalarMap):
    """
    
    :param descripcion_garantia:
    """
    
    def __init__(
            self,
            descripcion_garantia: str,
    ): 
        super().__init__({
            'DescripcionGarantia': descripcion_garantia,
        })
        

class DatosInmuebleType(ScalarMap):
    """
    
    :param tipo_inmueble:
    :param valor_referencia:
    :param codigo_postal:
    :param folio_real:
    """
    
    def __init__(
            self,
            tipo_inmueble: str,
            valor_referencia: str,
            codigo_postal: str,
            folio_real: str,
    ): 
        super().__init__({
            'TipoInmueble': tipo_inmueble,
            'ValorReferencia': valor_referencia,
            'CodigoPostal': codigo_postal,
            'FolioReal': folio_real,
        })
        

class DatosBienMutuoType(ScalarMap):
    """
    
    :param datos_inmueble:
    :param datos_otro:
    """
    
    def __init__(
            self,
            datos_inmueble: DatosInmuebleType | dict = None,
            datos_otro: DatosOtroType | dict = None,
    ): 
        super().__init__({
            'DatosInmueble': datos_inmueble,
            'DatosOtro': datos_otro,
        })
        

class DatosGarantiaType(ScalarMap):
    """
    
    :param tipo_garantia:
    :param datos_bien_mutuo:
    :param tipo_persona:
    """
    
    def __init__(
            self,
            tipo_garantia: str,
            datos_bien_mutuo: DatosBienMutuoType | dict = None,
            tipo_persona: TipoGaranteType | dict = None,
    ): 
        super().__init__({
            'TipoGarantia': tipo_garantia,
            'DatosBienMutuo': datos_bien_mutuo,
            'TipoPersona': tipo_persona,
        })
        

class DatosOperacionType(ScalarMap):
    """
    
    :param fecha_operacion:
    :param codigo_postal:
    :param tipo_operacion:
    :param datos_liquidacion:
    :param datos_garantia:
    """
    
    def __init__(
            self,
            fecha_operacion: date,
            codigo_postal: str,
            tipo_operacion: str,
            datos_liquidacion: DatosLiquidacionType | dict | Sequence[DatosLiquidacionType | dict],
            datos_garantia: DatosGarantiaType | dict | Sequence[DatosGarantiaType | dict] = None,
    ): 
        super().__init__({
            'FechaOperacion': fecha_operacion,
            'CodigoPostal': codigo_postal,
            'TipoOperacion': tipo_operacion,
            'DatosLiquidacion': datos_liquidacion,
            'DatosGarantia': datos_garantia,
        })
        

class DetalleOperacionesType(ScalarMap):
    """
    
    :param datos_operacion:
    """
    
    def __init__(
            self,
            datos_operacion: DatosOperacionType | dict | Sequence[DatosOperacionType | dict],
    ): 
        super().__init__({
            'DatosOperacion': datos_operacion,
        })
        

class FideicomisoSimpleType(ScalarMap):
    """
    
    :param denominacion_razon:
    :param rfc:
    :param identificador_fideicomiso:
    """
    
    def __init__(
            self,
            denominacion_razon: str,
            rfc: str = None,
            identificador_fideicomiso: str = None,
    ): 
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'Rfc': rfc,
            'IdentificadorFideicomiso': identificador_fideicomiso,
        })
        

class PersonaMoralSimpleType(ScalarMap):
    """
    
    :param denominacion_razon:
    :param fecha_constitucion:
    :param rfc:
    :param pais_nacionalidad:
    """
    
    def __init__(
            self,
            denominacion_razon: str,
            fecha_constitucion: date = None,
            rfc: str = None,
            pais_nacionalidad: str = None,
    ): 
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'FechaConstitucion': fecha_constitucion,
            'Rfc': rfc,
            'PaisNacionalidad': pais_nacionalidad,
        })
        

class PersonaFisicaSimpleType(ScalarMap):
    """
    
    :param nombre:
    :param apellido_paterno:
    :param apellido_materno:
    :param fecha_nacimiento:
    :param rfc:
    :param curp:
    :param pais_nacionalidad:
    """
    
    def __init__(
            self,
            nombre: str,
            apellido_paterno: str,
            apellido_materno: str,
            fecha_nacimiento: date = None,
            rfc: str = None,
            curp: str = None,
            pais_nacionalidad: str = None,
    ): 
        super().__init__({
            'Nombre': nombre,
            'ApellidoPaterno': apellido_paterno,
            'ApellidoMaterno': apellido_materno,
            'FechaNacimiento': fecha_nacimiento,
            'Rfc': rfc,
            'Curp': curp,
            'PaisNacionalidad': pais_nacionalidad,
        })
        

class TipoPersonaSimpleType(ScalarMap):
    """
    
    :param persona_fisica:
    :param persona_moral:
    :param fideicomiso:
    """
    
    def __init__(
            self,
            persona_fisica: PersonaFisicaSimpleType | dict = None,
            persona_moral: PersonaMoralSimpleType | dict = None,
            fideicomiso: FideicomisoSimpleType | dict = None,
    ): 
        super().__init__({
            'PersonaFisica': persona_fisica,
            'PersonaMoral': persona_moral,
            'Fideicomiso': fideicomiso,
        })
        

class DuenoBeneficiarioType(ScalarMap):
    """
    
    :param tipo_persona:
    """
    
    def __init__(
            self,
            tipo_persona: TipoPersonaSimpleType | dict,
    ): 
        super().__init__({
            'TipoPersona': tipo_persona,
        })
        

class TelefonoType(ScalarMap):
    """
    
    :param clave_pais:
    :param numero_telefono:
    :param correo_electronico:
    """
    
    def __init__(
            self,
            clave_pais: str = None,
            numero_telefono: str = None,
            correo_electronico: str = None,
    ): 
        super().__init__({
            'ClavePais': clave_pais,
            'NumeroTelefono': numero_telefono,
            'CorreoElectronico': correo_electronico,
        })
        

class ExtranjeroType(ScalarMap):
    """
    
    :param pais:
    :param estado_provincia:
    :param ciudad_poblacion:
    :param colonia:
    :param calle:
    :param numero_exterior:
    :param codigo_postal:
    :param numero_interior:
    """
    
    def __init__(
            self,
            pais: str,
            estado_provincia: str,
            ciudad_poblacion: str,
            colonia: str,
            calle: str,
            numero_exterior: str,
            codigo_postal: str,
            numero_interior: str = None,
    ): 
        super().__init__({
            'Pais': pais,
            'EstadoProvincia': estado_provincia,
            'CiudadPoblacion': ciudad_poblacion,
            'Colonia': colonia,
            'Calle': calle,
            'NumeroExterior': numero_exterior,
            'CodigoPostal': codigo_postal,
            'NumeroInterior': numero_interior,
        })
        

class NacionalType(ScalarMap):
    """
    
    :param colonia:
    :param calle:
    :param numero_exterior:
    :param codigo_postal:
    :param numero_interior:
    """
    
    def __init__(
            self,
            colonia: str,
            calle: str,
            numero_exterior: str,
            codigo_postal: str,
            numero_interior: str = None,
    ): 
        super().__init__({
            'Colonia': colonia,
            'Calle': calle,
            'NumeroExterior': numero_exterior,
            'CodigoPostal': codigo_postal,
            'NumeroInterior': numero_interior,
        })
        

class TipoDomicilioType(ScalarMap):
    """
    
    :param nacional:
    :param extranjero:
    """
    
    def __init__(
            self,
            nacional: NacionalType | dict = None,
            extranjero: ExtranjeroType | dict = None,
    ): 
        super().__init__({
            'Nacional': nacional,
            'Extranjero': extranjero,
        })
        

class RepresentanteApoderadoType(ScalarMap):
    """
    
    :param nombre:
    :param apellido_paterno:
    :param apellido_materno:
    :param fecha_nacimiento:
    :param rfc:
    :param curp:
    """
    
    def __init__(
            self,
            nombre: str,
            apellido_paterno: str,
            apellido_materno: str,
            fecha_nacimiento: date = None,
            rfc: str = None,
            curp: str = None,
    ): 
        super().__init__({
            'Nombre': nombre,
            'ApellidoPaterno': apellido_paterno,
            'ApellidoMaterno': apellido_materno,
            'FechaNacimiento': fecha_nacimiento,
            'Rfc': rfc,
            'Curp': curp,
        })
        

class FideicomisoType(ScalarMap):
    """
    
    :param denominacion_razon:
    :param apoderado_delegado:
    :param rfc:
    :param identificador_fideicomiso:
    """
    
    def __init__(
            self,
            denominacion_razon: str,
            apoderado_delegado: RepresentanteApoderadoType | dict,
            rfc: str = None,
            identificador_fideicomiso: str = None,
    ): 
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'ApoderadoDelegado': apoderado_delegado,
            'Rfc': rfc,
            'IdentificadorFideicomiso': identificador_fideicomiso,
        })
        

class PersonaMoralType(ScalarMap):
    """
    
    :param denominacion_razon:
    :param pais_nacionalidad:
    :param giro_mercantil:
    :param representante_apoderado:
    :param fecha_constitucion:
    :param rfc:
    """
    
    def __init__(
            self,
            denominacion_razon: str,
            pais_nacionalidad: str,
            giro_mercantil: str,
            representante_apoderado: RepresentanteApoderadoType | dict,
            fecha_constitucion: date = None,
            rfc: str = None,
    ): 
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'PaisNacionalidad': pais_nacionalidad,
            'GiroMercantil': giro_mercantil,
            'RepresentanteApoderado': representante_apoderado,
            'FechaConstitucion': fecha_constitucion,
            'Rfc': rfc,
        })
        

class PersonaFisicaType(ScalarMap):
    """
    
    :param nombre:
    :param apellido_paterno:
    :param apellido_materno:
    :param pais_nacionalidad:
    :param actividad_economica:
    :param fecha_nacimiento:
    :param rfc:
    :param curp:
    """
    
    def __init__(
            self,
            nombre: str,
            apellido_paterno: str,
            apellido_materno: str,
            pais_nacionalidad: str,
            actividad_economica: str,
            fecha_nacimiento: date = None,
            rfc: str = None,
            curp: str = None,
    ): 
        super().__init__({
            'Nombre': nombre,
            'ApellidoPaterno': apellido_paterno,
            'ApellidoMaterno': apellido_materno,
            'PaisNacionalidad': pais_nacionalidad,
            'ActividadEconomica': actividad_economica,
            'FechaNacimiento': fecha_nacimiento,
            'Rfc': rfc,
            'Curp': curp,
        })
        

class TipoPersonaType(ScalarMap):
    """
    
    :param persona_fisica:
    :param persona_moral:
    :param fideicomiso:
    """
    
    def __init__(
            self,
            persona_fisica: PersonaFisicaType | dict = None,
            persona_moral: PersonaMoralType | dict = None,
            fideicomiso: FideicomisoType | dict = None,
    ): 
        super().__init__({
            'PersonaFisica': persona_fisica,
            'PersonaMoral': persona_moral,
            'Fideicomiso': fideicomiso,
        })
        

class PersonaAvisoType(ScalarMap):
    """
    
    :param tipo_persona:
    :param tipo_domicilio:
    :param telefono:
    """
    
    def __init__(
            self,
            tipo_persona: TipoPersonaType | dict,
            tipo_domicilio: TipoDomicilioType | dict = None,
            telefono: TelefonoType | dict = None,
    ): 
        super().__init__({
            'TipoPersona': tipo_persona,
            'TipoDomicilio': tipo_domicilio,
            'Telefono': telefono,
        })
        

class AlertaType(ScalarMap):
    """
    
    :param tipo_alerta:
    :param descripcion_alerta:
    """
    
    def __init__(
            self,
            tipo_alerta: str,
            descripcion_alerta: str = None,
    ): 
        super().__init__({
            'TipoAlerta': tipo_alerta,
            'DescripcionAlerta': descripcion_alerta,
        })
        

class ModificatorioType(ScalarMap):
    """
    
    :param folio_modificacion:
    :param descripcion_modificacion:
    """
    
    def __init__(
            self,
            folio_modificacion: str,
            descripcion_modificacion: str,
    ): 
        super().__init__({
            'FolioModificacion': folio_modificacion,
            'DescripcionModificacion': descripcion_modificacion,
        })
        

class AvisoType(ScalarMap):
    """
    
    :param referencia_aviso:
    :param prioridad:
    :param alerta:
    :param persona_aviso:
    :param detalle_operaciones:
    :param modificatorio:
    :param dueno_beneficiario:
    """
    
    def __init__(
            self,
            referencia_aviso: str,
            prioridad: str,
            alerta: AlertaType | dict,
            persona_aviso: PersonaAvisoType | dict | Sequence[PersonaAvisoType | dict],
            detalle_operaciones: DetalleOperacionesType | dict,
            modificatorio: ModificatorioType | dict = None,
            dueno_beneficiario: DuenoBeneficiarioType | dict | Sequence[DuenoBeneficiarioType | dict] = None,
    ): 
        super().__init__({
            'ReferenciaAviso': referencia_aviso,
            'Prioridad': prioridad,
            'Alerta': alerta,
            'PersonaAviso': persona_aviso,
            'DetalleOperaciones': detalle_operaciones,
            'Modificatorio': modificatorio,
            'DuenoBeneficiario': dueno_beneficiario,
        })
        

class SujetoObligadoType(ScalarMap):
    """
    
    :param clave_sujeto_obligado:
    :param clave_actividad:
    :param clave_entidad_colegiada:
    :param exento:
    """
    
    def __init__(
            self,
            clave_sujeto_obligado: str,
            clave_actividad: str,
            clave_entidad_colegiada: str = None,
            exento: str = None,
    ): 
        super().__init__({
            'ClaveSujetoObligado': clave_sujeto_obligado,
            'ClaveActividad': clave_actividad,
            'ClaveEntidadColegiada': clave_entidad_colegiada,
            'Exento': exento,
        })
        

class InformeType(ScalarMap):
    """
    
    :param mes_reportado:
    :param sujeto_obligado:
    :param aviso:
    """
    
    def __init__(
            self,
            mes_reportado: str,
            sujeto_obligado: SujetoObligadoType | dict,
            aviso: AvisoType | dict | Sequence[AvisoType | dict] = None,
    ): 
        super().__init__({
            'MesReportado': mes_reportado,
            'SujetoObligado': sujeto_obligado,
            'Aviso': aviso,
        })
        

class ArchivoType(ScalarMap):
    """
    
    :param informe:
    """
    
    def __init__(
            self,
            informe: InformeType | dict | Sequence[InformeType | dict],
    ): 
        super().__init__({
            'Informe': informe,
        })
        

class Archivo(ArchivoType, XElement):
    tag = '{http://www.uif.shcp.gob.mx/recepcion/mpc}archivo'




import logging
import os
from datetime import date, datetime
from itertools import chain
from unittest import mock

import pytest
from OpenSSL import crypto

from satcfdi.models.certificate import CertificateType
from satcfdi.models.certificate_request import CertificateSigningRequest
from satcfdi.exceptions import CFDIError
from satcfdi.models.curp import CURP
from satcfdi.models.date_period import DatePeriod
from satcfdi.models.rfc import RFC, RFCType, RFC_Generico_Nacional, RFC_Generico_Extranjero
from satcfdi.models.signer import Signer
from satcfdi.pacs.sat import SAT
from satcfdi.transform import verify_certificate
from tests.constants import PERSONAS_FISICAS, PERSONAS_MORALES
from tests.utils import get_signer, SAT_Certificate_Store_Pruebas

module = 'satcfdi'
current_dir = os.path.dirname(__file__)

logging.basicConfig(level=logging.INFO)

sat = SAT()


def test_descarga_certificado_sat_pac():
    cert = sat.recover_certificate("00001000000504465028")

    # assert cert.issuer().startswith("1.2.840.113549.1.9.2=responsable: ADMINISTRACION CENTRAL DE SERVICIOS TRIBUTARIOS AL CONTRIBUYE")
    # assert cert.subject().startswith(r"OU=MEGAPACSAT970701NN301,2.5.4.5=\ / GARJ750416HDFRVR09,2.5.4.45=SAT970701NN3 / GARJ7504168N4,O=SERVIC")
    # assert cert.issuer().startswith("CN=AUTORIDAD CERTIFICADORA,O=SERVICIO DE ADMINISTRACION TRIBUTARIA,OU=SAT-IES Authority,emailAddress=contacto.tecnic")
    # assert cert.subject().startswith("CN=SERVICIO DE ADMINISTRACION TRIBUTARIA,name=SERVICIO DE ADMINISTRACION TRIBUTARIA,O=SERVICIO DE ADMINISTRACION TRIBUTARIA,x500Unique")
    assert cert.certificate_base64().startswith("MIIGKTCCBBGgAwIBAgIUMDAwMDEwMDAwMDA1MDQ0NjUwM")
    assert str(cert.serial_number) == '275106190557734483187066766829380167334928331320'
    assert cert.certificate_number == "00001000000504465028"
    assert cert.branch_name == "MEGAPACSAT970701NN301"
    assert cert.rfc_pac == "SAT970701NN3"
    assert cert.rfc_representante == "GARJ7504168N4"
    assert cert.curp is None
    assert cert.curp_representante == "GARJ750416HDFRVR09"
    assert str(cert.serial_number) == "275106190557734483187066766829380167334928331320"
    assert cert.rfc_pac == "SAT970701NN3"
    assert cert.rfc_representante == "GARJ7504168N4"
    assert cert.legal_name == 'SERVICIO DE ADMINISTRACION TRIBUTARIA'
    assert cert.email is None


def test_signer_razon_social():
    res = {
        'CACX7605101P8': 'XOCHILT CASAS CHAVEZ',
        'FUNK671228PH6': 'KARLA FUENTE NOLASCO',
        'IAÑL750210963': 'LUIS IAN ÑUZCO',
        'JUFA7608212V6': 'ADRIANA JUAREZ FERNANDEZ',
        'KAHO641101B39': 'OSCAR KALA HAAK',
        'KICR630120NX3': 'RODRIGO KITIA CASTRO',
        'MISC491214B86': 'CECILIA MIRANDA SANCHEZ',
        'RAQÑ7701212M3': 'ÑEVES RAMIREZ QUEZADA',
        'WATM640917J45': 'MARIA WATEMBER TORRES',
        'WERX631016S30': 'XAIME WEIR ROJO',
        'XAMA620210DQ5': 'ALBA XKARAJAM MENDEZ',
        'XIQB891116QE4': 'BERENICE XIMO QUEZADA',
        'XOJI740919U48': 'INGRID XODAR JIMENEZ',
        'EKU9003173C9': 'ESCUELA KEMPER URGATE',
        'EWE1709045U0': 'ESCUELA WILSON ESQUIVEL',
        'H&E951128469': 'HERRERIA & ELECTRICOS',
        'HAÑ930228SM9': 'HERMANOS ANZURES ÑARVAEZ',
        'IIA040805DZ4': 'INDISTRIA ILUMINADORA DE ALMACENES',
        'IVD920810GU2': 'INNOVACION VALOR Y DESARROLLO',
        'IXS7607092R5': 'INTERNACIONAL XIMBO Y SABORES',
        'JES900109Q90': 'JIMENEZ ESTRADA SALAS A A',
        'KIJ0906199R1': 'KERNEL INDUSTIA JUGUETERA',
        'L&O950913MSA': 'LUCES & OBRAS',
        'OÑO120726RX3': 'ORGANICOS ÑAVEZ OSORIO S.A DE C.V',
        'S&S051221SE2': 'S & SOFTWARE',
        'URE180429TM6': 'UNIVERSIDAD ROBOTICA ESPAÑOLA',
        'XIA190128J61': 'XENON INDUSTRIAL ARTICLES',
        'ZUÑ920208KL4': 'ZAPATERIA URTADO ÑERI'
    }

    for rfc in chain(PERSONAS_FISICAS, PERSONAS_MORALES):
        try:
            signer = get_signer(rfc, get_csd=True)
            rs = signer.legal_name
            assert res[rfc] == rs
            assert rfc == signer.rfc
        except FileNotFoundError as ex:
            pass


def test_curp():
    curp = CURP("FEPM890204HASRRN08")
    assert curp.is_valid()
    assert curp.estado == "AS"

    # for rfc in PERSONAS_FISICAS:
    #     signer = get_signer(rfc)
    #     assert not signer.curp.is_valid()


def test_crypto_razon_social():
    for rfc in chain(PERSONAS_FISICAS, PERSONAS_MORALES):
        try:
            signer = get_signer(rfc, get_csd=True)

            cer = crypto.load_certificate(crypto.FILETYPE_ASN1, signer.certificate_bytes())
            subj = cer.get_subject()
            for k, v in subj.get_components():
                k.decode("latin_1")
                v.decode("latin_1")

        except (FileNotFoundError, ValueError) as ex:
            pass


def test_verify_certificates():
    for rfc in chain(PERSONAS_FISICAS, PERSONAS_MORALES):
        with mock.patch(f'{module}.transform.SAT_Certificate_Store', SAT_Certificate_Store_Pruebas):
            try:
                signer = get_signer(rfc)
                verify_certificate(signer, at=datetime(2021, 6, 12))
                assert signer.type == CertificateType.Fiel

                signer_csd = get_signer(rfc, get_csd=True)
                verify_certificate(signer_csd, at=datetime(2021, 6, 12))
                assert signer_csd.type == CertificateType.CSD

            except FileNotFoundError as ex:
                pass


def test_requirement():
    for rfc in PERSONAS_FISICAS:
        res = CertificateSigningRequest.load(
            request=open(os.path.join(current_dir, "csd", "Personas Fisicas", f'{rfc.lower()}.req'), 'rb').read()
        )
    for rfc in PERSONAS_MORALES:
        res = CertificateSigningRequest.load(
            request=open(os.path.join(current_dir, "csd", "Personas Morales", f'{rfc.lower()}.req'), 'rb').read()
        )


def test_thumbprint():
    signer = get_signer("xiqb891116qe4")

    assert signer.fingerprint().hex() == 'cad8b0aa7940aec665d9db9755b995b8631409c4'

    signer2 = get_signer("h&e951128469")

    with pytest.raises(CFDIError) as ex:
        signer = Signer(
            key=signer.key,
            certificate=signer2.certificate,
            check=True
        )

    password = "IGUESSTHISISSECURE"
    p12 = signer.pcks12_bytes(password)
    with open(os.path.join(current_dir, "csd", f'xiqb891116qe4.pfx'), 'wb') as f:
        f.write(p12)

    pkcs12_signer = Signer.load_pkcs12(
        data=open(os.path.join(current_dir, "csd", f'xiqb891116qe4.pfx'), 'rb').read(),
        password=password
    )

    assert pkcs12_signer.legal_name == 'BERENICE XIMO QUEZADA'

    os.remove(os.path.join(current_dir, "csd", f'xiqb891116qe4.pfx'))


def test_signer_change_password():
    rfc = 'xiqb891116qe4'
    signer = get_signer(rfc)

    password = "THISISMORESCURE"
    key = signer.key_bytes(password)
    cer = signer.certificate_bytes()

    with open(os.path.join(current_dir, "csd", 'xiqb891116qe4.out.key'), 'wb') as f:
        f.write(key)
    with open(os.path.join(current_dir, "csd", 'xiqb891116qe4.out.cer'), 'wb') as f:
        f.write(cer)
    with open(os.path.join(current_dir, "csd", 'xiqb891116qe4.out.txt'), 'w') as f:
        f.write(password)

    signer = Signer.load(
        certificate=open(os.path.join(current_dir, f'csd/{rfc}.out.cer'), 'rb').read(),
        key=open(os.path.join(current_dir, f'csd/{rfc}.out.key'), 'rb').read(),
        password=open(os.path.join(current_dir, f'csd/{rfc}.out.txt'), 'rb').read()
    )

    os.remove(os.path.join(current_dir, "csd", 'xiqb891116qe4.out.key'))
    os.remove(os.path.join(current_dir, "csd", 'xiqb891116qe4.out.cer'))
    os.remove(os.path.join(current_dir, "csd", 'xiqb891116qe4.out.txt'))


def test_rfc():
    for rfc in PERSONAS_FISICAS:
        r = RFC(rfc, entity_type=RFCType.FISICA)
        assert r.is_valid()

    for rfc in PERSONAS_MORALES:
        r = RFC(rfc, entity_type=RFCType.MORAL)
        assert r.is_valid()

    r = RFC_Generico_Nacional
    assert r.is_valid()
    assert r.type == RFCType.FISICA

    r = RFC_Generico_Extranjero
    assert r.is_valid()
    assert r.type == RFCType.FISICA


def test_period():
    period = DatePeriod(year=2020, month=4)
    my_date = date(year=2021, month=3, day=20)
    assert my_date > period
    assert my_date >= period
    assert my_date != period
    assert not my_date == period
    assert not my_date <= period
    assert not my_date < period

    assert str(period) == '2020-04'

    period = DatePeriod(year=None)
    my_date = date(year=2021, month=3, day=20)
    assert my_date >= period
    assert my_date == period
    assert my_date <= period
    assert not my_date > period
    assert not my_date < period
    assert not my_date != period

    assert str(period) == ""

    assert not period == 1
    assert period != 1
    with pytest.raises(TypeError):
        assert not period < 1
    with pytest.raises(TypeError):
        assert not period > 1
    with pytest.raises(TypeError):
        assert not period >= 1
    with pytest.raises(TypeError):
        assert not period <= 1
